from datetime import datetime  # noqa


class TipReportFilterDto(object):
    def __init__(self, start_date=None, end_date=None, operator_id=None):
        # type: (datetime, datetime, str) -> None
        self.start_date = start_date
        self.end_date = end_date
        self.operator_id = operator_id



import logging

from model.customexception import ProductNotExistsException
from typing import Optional

from actions.products.get_products import (
    get_product_by_barcode,
    get_product_price_from_cache,
)
from sysactions import (
    get_model,
    get_podtype,
    show_messagebox,
)
from systools import sys_log_exception

logger = logging.getLogger("PosActions")


def sell_product_by_barcode(
    pos_id,                   # type: str
    barcode_number,           # type: str
    check_consistency=False,  # type: Optional[bool]
    received_qty="1",         # type: Optional[str]
):
    # type: (...) -> str

    try:
        from posactions import doSale, doCompleteOption, doChangeQuantity
        if not barcode_number:
            return "False"

        if check_consistency and not _is_a_valid_ean13(barcode=barcode_number):
            logger.info("Barcode {0} is not a Valid EAN13".format(barcode_number))
            ret = show_messagebox(pos_id, "$BARCODE_WRONG", buttons="$OK|$CANCEL")
            if ret == 0:
                return "Clear"
            return "False"

        if str(barcode_number).zfill(13).startswith('2'):
            logger.info("Barcode {0} is a Scale Type Code".format(barcode_number))
            plu, calc_qty = _ticket_from_scale(str(barcode_number), pos_id)
            doCompleteOption(str(pos_id), "1", plu, calc_qty)
        else:
            logger.info("Barcode {0} ".format(barcode_number))
            product_by_barcode = get_product_by_barcode(barcode_number)
            if product_by_barcode:
                if product_by_barcode["ruptured"]:
                    logger.info("Scanned barcode {0} product is ruptured!".format(barcode_number))
                    user_message = "$BARCODE_PRODUCT_RUPTURED|{}".format(product_by_barcode["name"])
                    ret = show_messagebox(pos_id, user_message, buttons="$YES|$NO")
                    if ret != 0:
                        return "Clear"

                doCompleteOption(str(pos_id), "1", product_by_barcode['plu'], received_qty)
            else:
                logger.info("Scanned barcode {0} not found on the product database!".format(barcode_number))
                ret = show_messagebox(pos_id, "$BARCODE_NOT_FOUND|{}".format(barcode_number), buttons="$OK|$CANCEL")
                if ret == 0:
                    return "Clear"
                return "False"
        return "True"
    except ProductNotExistsException:
        return "Clear"
    except Exception as _:
        sys_log_exception("Error on [sell_product_by_barcode]")
        logger.exception("Error on [sell_product_by_barcode]")
        return "False"


def _is_a_valid_ean13(
    barcode,   # type: str
):
    # type: (...) -> bool

    if len(str(barcode)) > 13:
        return False
    
    barcode = list(str(barcode).zfill(13))
    index = 0
    value = 0
    dv = barcode.pop()
    for x in barcode:
        value = value + (int(x) * ((index % 2) * 2 + 1))
        index += 1
    value2 = (((int(value / 10) + 1) * 10) - value) % 10
    if value2 != int(dv):
        return False
    return True


def _ticket_from_scale(
    barcode,   # type: str
    pos_id,    # type: str
):
    # type: (...) -> (str, float)

    from posactions import (
        scale_ticket_mode,
        barcode_pattern_reading,
    )

    try:
        first_x = barcode_pattern_reading.find("X")
        first_0 = barcode_pattern_reading.find("0")
        first_y = barcode_pattern_reading.find("Y")
        first_w = barcode_pattern_reading.find("W")
        plu = int(barcode[first_x:first_0 if first_0 > -1 else first_y])
        param = int(barcode[first_y:first_w])

        if scale_ticket_mode == 'price':
            final_price = float(param) / 100
            price = float(get_product_price_from_cache(plu, pos_id))
            qtd = round((final_price / price), 3)
        else:
            qtd = round(float(param) / 1000, 3)

        return str(plu), qtd
    except ProductNotExistsException as ex:
        logger.info("The product {} in Barcode does not exists".format(ex.message))
        show_messagebox(pos_id, "$PRODUCT_DOES_NOT_EXISTS", "$ERROR")
        raise



import base64
from logging import Logger
from threading import Lock

from model.error_type import ExternalServicesErrorsType

from interactor.route.order.finalize.dto import (
    OrderFinalizeRequest,
    FiscalOrderResponse,
)
from model.config import Config
from model.exception import ExternalServiceErrors
from model.repository import PosRepositoryModel
from services import FiscalService, PrinterService
from validation_util import (
    check_if_business_day_is_opened,
    check_if_operator_is_logged,
    check_if_order_is_totaled,
    validate_api_key,
    check_if_request_order_id_is_equal_current_order_id,
    check_if_state_and_due_amount_is_valid_to_finalize,
)

order_finalize_lock = {}


class OrderFinalizeInteractor(object):
    def __init__(
        self,
        logger,              # type: Logger
        config,              # type: Config
        pos_repository,      # type: PosRepositoryModel
        fiscal_service,      # type: FiscalService
        printer_service,     # type: PrinterService
    ):
        # type: (...) -> None

        self.logger = logger
        self.config = config
        self.pos_repository = pos_repository
        self.fiscal_service = fiscal_service
        self.printer_service = printer_service

    def execute(
        self,
        request,    # type: OrderFinalizeRequest
    ):
        # type: (...) -> FiscalOrderResponse

        pos_id = validate_api_key(request, self.config.api_key_mapping)
        order_id = request.order_id

        self.logger.info("Executing finalize order: [{}-{}]".format(request.pos_id, order_id))
        with self._get_pos_lock(pos_id):
            self.logger.info("Locking POS to process order: [{}-{}]".format(request.pos_id, order_id))

            model = self.pos_repository.get_model(pos_id)

            check_if_request_order_id_is_equal_current_order_id(
                model=model,
                request_order_id=order_id,
            )
            check_if_business_day_is_opened(self.pos_repository, model)
            check_if_operator_is_logged(self.pos_repository, model)
            check_if_state_and_due_amount_is_valid_to_finalize(
                logger=self.logger,
                pos_id=request.pos_id,
                order_id=order_id,
                model=model,
                pos_repository=self.pos_repository,
            )

            pos_ot = self.pos_repository.get_pos_ot(model)
            try:
                self.logger.info("Processing order fiscal: [{}-{}]".format(request.pos_id, order_id))
                response = self.fiscal_service.process_order_fiscal(
                    model=model,
                    print_coupon=request.print_coupon,
                    device_type=request.device_type,
                )
            except Exception as _:
                self.pos_repository.void_tenders_and_order(pos_id, pos_ot)
                self.logger.exception("Error fiscalizing order. The order has been voided")
                raise ExternalServiceErrors(ExternalServicesErrorsType.FiscalizeUnavailable)

            self.logger.info("Set order to paid: [{}-{}]".format(request.pos_id, order_id))
            self.pos_repository.set_order_to_paid(
                pos_id=pos_id,
                order_id=order_id,
                model=model,
                pos_ot=pos_ot,
            )

            self.logger.info("Get order customer receipts: [{}-{}]".format(request.pos_id, order_id))
            customer_receipts = self.fiscal_service.get_customer_receipts(
                model=model,
            )

            if request.print_tef_customer_receipt:
                self.logger.info("Send order to print: [{}-{}]".format(request.pos_id, order_id))
                for tef_customer_receipt in customer_receipts:
                    self.printer_service.send_print_data(
                        model=model,
                        response=base64.b64decode(tef_customer_receipt),
                    )

            self.logger.info("Release lock order: [{}-{}]".format(request.pos_id, order_id))
            return FiscalOrderResponse(
                fiscal_coupon=base64.b64encode(response) if response else "",
                tef_customer_receipt=customer_receipts,
            )

    @staticmethod
    def _get_pos_lock(
            pos_id,     # type: str
    ):
        # type: (...) -> Lock

        global order_finalize_lock

        if pos_id not in order_finalize_lock:
            order_finalize_lock[pos_id] = Lock()

        return order_finalize_lock[pos_id]



from mw_helper import ensure_iterable

from ._OrderCondition import OrderCondition
from production.model import ProductionOrder


class CustomPropertyValuesCondition(OrderCondition):
    def __init__(self, custom_property, allowed_values, excluded_values):
        self.custom_property = custom_property
        self.allowed_values = ensure_iterable(allowed_values or [])
        self.excluded_values = ensure_iterable(excluded_values or [])

    def evaluate(self, order):
        # type: (ProductionOrder) -> bool

        if self.custom_property in order.properties:
            property_value = order.properties[self.custom_property]
            if self.allowed_values and property_value in self.allowed_values:
                return True
            if self.excluded_values and property_value not in self.excluded_values:
                return True
        return False



from typing import Optional
from xml.etree import cElementTree as eTree

from sysactions import (
    StopAction,
    show_messagebox,
    get_model,
    get_posot,
    get_current_order,
)

from actions.order import synchronize_ui


def must_sync_ui(
    pos_id,             # type: str
    different_than,     # type: str
):
    # type: (...) -> None

    order = _get_order_picture(pos_id=pos_id)

    different_than = different_than.upper()
    if order.get("state") != different_than:
        show_messagebox(
            posid=pos_id,
            message="$ERROR_ORDER_{}_NOT_SYNCHRONIZED".format(different_than),
            buttons="$OK",
        )

        synchronize_ui(
            pos_id=pos_id,
        )

        raise StopAction()


def _get_order_picture(
    pos_id,     # type: Optional[str]
):
    # type: (...) -> Optional[eTree.Element]

    model = get_model(posid=pos_id)
    pos_ot = get_posot(model=model)

    current_order = get_current_order(model=model)
    order_id = current_order.get("orderId")
    order_xml = eTree.XML(
        pos_ot.orderPicture(
            posid=pos_id,
            orderid=order_id,
        )
    )
    return order_xml.find("Order") if order_xml.find("Order") else order_xml



from production.box import JitItemFilterBox
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class JitItemFilterBoxBuilder(BoxBuilder):
    def __init__(self, product_repository, loggers):
        super(JitItemFilterBoxBuilder, self).__init__(loggers)
        self.product_repository = product_repository

    def build(self, config):
        # type: (BoxConfiguration) -> JitItemFilterBox

        forbidden_jit_lines = self.get_extra(config, "ForbiddenJitLines", None)
        return JitItemFilterBox(
            config.name,
            config.sons,
            self.product_repository,
            self.get_extra(config, "AllowedJitLines", None),
            forbidden_jit_lines,
            self.get_logger(config)
        )



import json

import requests

from requests import Response
from typing import Optional, List, Tuple
from xml.etree import cElementTree as eTree

from msgbus import (
    FM_PARAM,
    TK_CMP_MESSAGE,
    TK_SLCTRL_OMGR_ORDERPICT,
    TK_SYS_NAK,
)
from messagebus import Message
from mbcontextmessagehandler import MbContextMessageBus

from helper import retry

from application.model.configuration import Configurations
from application.model.customexception import (
    NotFound,
    ForbiddenError,
    ProcessorError,
    InUseError,
    NotAvailableError,
    NotFoundError,
    NoResponseError,
    AlreadyUnlockedError,
    NotAllowedError,
    ClientCustomError,
)

from application.model.orderinfo import Order, OrderJsonEncoder
from util import convert_xml_to_dict


class APIRepository(object):

    def __init__(
        self,
        configs,          # type: Configurations
        message_bus,      # type: MbContextMessageBus
    ):
        # type: (...) -> None

        self.logger = configs.logger
        self.configs = configs
        self.message_bus = message_bus

    def retrieve_loyalty_customer_info(self, loyalty_customer_id):
        # type: (str) -> Optional[str, Exception]
        try:
            url = self._get_customer_url(loyalty_customer_id)
            headers = self._get_headers()
            body = dict(storeId=self.configs.store_id)
            errors_list = []

            response = retry(
                logger=self.logger,
                method_name="retrieve_loyalty_customer_info",
                n=1,
                sleep_seconds=1,
                trying_to_do=lambda: requests.request(
                    method="GET",
                    url=url,
                    json=body,
                    headers=headers,
                    timeout=5,
                ),
                is_success=lambda resp: resp.status_code == 200,
                success=lambda resp: resp,
                failed=lambda resp: errors_list.append((resp.status_code, resp))
            )
            if response:
                return response.content

            if not errors_list:
                raise NotFound()

            response_error = self._get_response_error(errors_list)
            self._raise_lock_custom_exception(response_error.status_code, response_error)
        except Exception as ex:
            self.logger.error("Error on get_benefit: {}".format(ex))
            raise

    def retrieve_benefit(
        self,
        pos_id,         # type: int
        voucher_id,     # type: str
    ):
        # type: (...) -> Optional[str, bool]

        order_info = self._get_order_info(pos_id=pos_id)
        benefit_raw = self._get_benefit(
            benefit_id=voucher_id,
            order=order_info,
        )

        if not benefit_raw:
            self.logger.info("No benefit found")
            return None

        benefit = json.loads(benefit_raw)
        benefit = benefit.get("benefit", benefit)

        notify = benefit.get("notify", False)
        benefit_id = benefit.get("id", None)
        if notify and benefit_id:
            self._lock_benefit(benefit_id=benefit_id)

        return benefit_raw

    def _get_table_info(
        self,
        pos_id,      # type: int
        table_id,    # type: int
    ):
        # type: (...) -> Optional[dict]

        if not table_id:
            return None

        params = ("TABLE_PICTURE", pos_id, table_id)
        req_buf = "\0".join([str(p) for p in params])
        message = Message(
            token=TK_CMP_MESSAGE,
            data_type=FM_PARAM,
            data=req_buf,
        )
        msg = self.message_bus.send_message(
            component='TABLEMGR',
            message=message,
        )

        if msg.token != TK_SYS_NAK:
            return convert_xml_to_dict(xml_object=eTree.XML(msg.data))
        return {}

    def _get_order_info(
        self,
        pos_id,         # type: int
        order_id="",    # type: Optional[str]
        no_cache="",    # type: Optional[str]
    ):
        # type: (...) -> dict

        params = (pos_id, order_id, 'true' if no_cache else 'false')
        req_buf = '\x00'.join((str(p) for p in params))

        message = Message(
            token=TK_SLCTRL_OMGR_ORDERPICT,
            data_type=FM_PARAM,
            data=req_buf,
        )
        msg = self.message_bus.send_message(
            component='ORDERMGR' + str(pos_id),
            message=message,
        )

        if msg.token == TK_SYS_NAK:
            return {}

        order_info = convert_xml_to_dict(xml_object=eTree.XML(msg.data.split('\0')[2]))
        custom_properties = order_info.get('CustomOrderProperties')
        table_id = custom_properties.get('TABLE_ID') if custom_properties and "TABLE_ID" in custom_properties else None
        table_info = self._get_table_info(
            pos_id=pos_id,
            table_id=table_id,
        )

        order = Order(
            store_id=self.configs.store_id,
            order_info=order_info,
            table_info=table_info,
        )
        return json.loads(json.dumps(order, encoding="utf-8", cls=OrderJsonEncoder))

    def burn_benefit(
        self,
        benefit_id,     # type: str
    ):
        # type: (...) -> None

        url = self._get_burn_url(benefit_id)
        headers = self._get_headers()
        errors_list = []

        response = retry(
            logger=self.logger,
            method_name="burn_benefit",
            n=1,
            sleep_seconds=1,
            trying_to_do=lambda: requests.request("PUT", url=url, headers=headers, timeout=20),
            is_success=lambda resp: resp.status_code == 204,
            success=lambda resp: resp,
            failed=lambda resp: errors_list.append((resp.status_code, resp))
        )
        if response:
            return response.content
        else:
            if not errors_list:
                raise NotFound()

            response_error = self._get_response_error(errors_list)
            self._raise_unlock_burn_custom_exception(response_error.status_code, response_error)

    def _get_benefit(
        self,
        benefit_id,     # type: str
        order,          # type: str
    ):
        # type: (...) -> str

        url = self._get_benefit_url(voucher_id=benefit_id)
        headers = self._get_headers()
        errors_list = []

        response = retry(
            logger=self.logger,
            method_name="get_benefit",
            n=1,
            sleep_seconds=1,
            trying_to_do=lambda: requests.request(
                "POST",
                url=url,
                headers=headers,
                timeout=20,
                json=order,
            ),
            is_success=lambda resp: resp.status_code == 200,
            success=lambda resp: resp,
            failed=lambda resp: errors_list.append((resp.status_code, resp))
        )

        if response:
            return response.content
        else:
            if not errors_list:
                raise NotFound()

            response_error = self._get_response_error(errors_list)
            self._raise_lock_custom_exception(response_error.status_code, response_error)

    def _lock_benefit(
        self,
        benefit_id  # type: str
    ):
        # type: (...) -> None

        url = self._get_lock_url(benefit_id=benefit_id)
        headers = self._get_headers()
        errors_list = []

        response = retry(
            logger=self.logger,
            method_name="lock_benefit",
            n=1,
            sleep_seconds=1,
            trying_to_do=lambda: requests.request("PUT", url=url, headers=headers, timeout=20),
            is_success=lambda resp: resp.status_code == 204,
            success=lambda resp: resp,
            failed=lambda resp: errors_list.append((resp.status_code, resp))
        )

        if response:
            return response.content
        else:
            if not errors_list:
                raise NotFound()

            response_error = self._get_response_error(errors_list=errors_list)
            self._raise_lock_custom_exception(
                status_code=response_error.status_code,
                response_error=response_error,
            )

    def unlock_benefit(
        self,
        benefit_id,     # type: str
    ):
        # type: (...) -> None

        url = self._get_unlock_url(benefit_id)
        headers = self._get_headers()
        errors_list = []

        response = retry(
            logger=self.logger,
            method_name="unlock_benefit",
            n=1,
            sleep_seconds=1,
            trying_to_do=lambda: requests.request("PUT", url=url, headers=headers, timeout=20),
            is_success=lambda resp: resp.status_code == 204,
            success=lambda resp: resp,
            failed=lambda resp: errors_list.append((resp.status_code, resp))
        )

        if response:
            return response.content
        else:
            if not errors_list:
                raise NotFound()

            response_error = self._get_response_error(errors_list)
            self._raise_unlock_burn_custom_exception(response_error.status_code, response_error)

    def _get_customer_url(self, loyalty_customer_id):
        # type: (str) -> str

        end_point_url = self.configs.endpoints.retrieve_loyalty_customer_info.format(customerId=loyalty_customer_id)
        full_url = self.configs.base_url + end_point_url

        self.logger.info("Requesting from get customer info endpoint: {}".format(full_url))

        return full_url

    def _get_benefit_url(self, voucher_id):
        # type: (str) -> str

        end_point_url = self.configs.endpoints.get_benefit.format(voucherId=voucher_id)
        full_url = self.configs.base_url + end_point_url

        self.logger.info("Requesting from get benefit endpoint: {}".format(full_url))

        return full_url

    def _get_burn_url(self, benefit_id):
        # type: (str) -> str

        end_point_url = self.configs.endpoints.burn_benefit.format(benefitId=benefit_id)
        full_url = self.configs.base_url + end_point_url

        self.logger.info("Requesting from burn endpoint: {}".format(full_url))

        return full_url

    def _get_lock_url(
        self,
        benefit_id  # type: str
    ):
        # type: (...) -> str

        end_point_url = self.configs.endpoints.lock_benefit.format(benefitId=benefit_id)
        full_url = self.configs.base_url + end_point_url

        self.logger.info("Requesting from lock endpoint: {}".format(full_url))

        return full_url

    def _get_unlock_url(self, benefit_id):
        # type: (str) -> str

        end_point_url = self.configs.endpoints.unlock_benefit.format(benefitId=benefit_id)
        full_url = self.configs.base_url + end_point_url

        self.logger.info("Requesting from unlock endpoint: {}".format(full_url))

        return full_url

    def _get_headers(self):
        # type: () -> dict

        return {"X-Api-Key": self.configs.api_key, "Content-type": "application/json"}

    def _raise_lock_custom_exception(self, status_code, response_error):
        # type: (int, Response) -> None

        if status_code == 204:
            raise NoResponseError(self.logger, response_error)
        if status_code == 403:
            raise ForbiddenError(self.logger, response_error)
        if status_code == 404:
            raise NotFoundError(self.logger, response_error)
        if status_code == 405:
            raise NotAllowedError(self.logger, response_error)
        if status_code == 406:
            raise ClientCustomError(self.logger, response_error)
        if status_code == 409:
            raise InUseError(self.logger, response_error)
        if status_code == 410:
            raise NotAvailableError(self.logger, response_error)
        if status_code >= 500:
            raise ProcessorError(self.logger, response_error)

    def _raise_unlock_burn_custom_exception(self, status_code, response_error):
        # type: (int, Response) -> None

        if status_code == 204:
            raise NoResponseError(self.logger, response_error)
        if status_code == 400:
            raise InUseError(self.logger, response_error)
        if status_code == 403:
            raise ForbiddenError(self.logger, response_error)
        if status_code == 404:
            raise NotFoundError(self.logger, response_error)
        if status_code == 405:
            raise NotAllowedError(self.logger, response_error)
        if status_code == 406:
            raise NotAvailableError(self.logger, response_error)
        if status_code == 409:
            raise AlreadyUnlockedError(self.logger, response_error)
        if status_code >= 500:
            raise ProcessorError(self.logger, response_error)

    @staticmethod
    def _get_response_error(errors_list):
        # type: (List[Tuple]) -> Response

        last_error = errors_list[-1]
        status_code, response = last_error
        return response



from production.model import EventNotFoundException
from production.treebuilder import ViewConfiguration
from ._CommonViewBuilder import get_logger
from production.view import PublishView


class PublishViewBuilder(object):
    def __init__(self, message_bus, loggers):
        self.message_bus = message_bus
        self.loggers = loggers

    def build(self, config):
        # type: (ViewConfiguration) -> PublishView

        if "Event" not in config.extra_config:
            raise EventNotFoundException("PublishView without Event: {}".format(config.name))

        return PublishView(config.name,
                           self.message_bus,
                           config.extra_config["Event"],
                           config.extra_config.get("EventType", "MultipleEvent"),
                           config.extra_config.get("Tag", ""),
                           bool(config.extra_config.get("SendOrder", "true").lower() == "true"),
                           get_logger(config, self.loggers))



import json
import logging

from xml.etree import cElementTree as eTree

from bustoken import TK_ELECTRONIC_PAYMENT_CANCEL_PAYMENT_ORDER
from msgbus import (
    FM_PARAM,
    TK_SYS_ACK,
)
from sysactions import (
    get_posot,
    mbcontext,
    show_messagebox,
)
from typing import (
    Optional,
    Tuple
)

from actions.eft import (
    get_electronic_payment_component_name,
    has_electronic_payment,
    get_electronic_tenders,
    save_cancel_receipt
)
from application.interactor import OrderInteractor

logger = logging.getLogger("EftActions")


def cancel_eft_transactions(
    pos_id,                     # type: str
    model,                      # type: eTree.XML
    order_id,                   # type: str
    tender=None,                # type: Optional[eTree.XML]
    is_table_service=False,     # type: bool
    is_mobile=False,            # type: bool
    operator_id=None,           # type: Optional[str]
    order=None,                 # type: eTree.Element
):
    # type: (...) -> bool

    if order is None:
        order_taker = get_posot(model)
        order_picture = order_taker.orderPicture(
            posid=pos_id,
            orderid=order_id or ""
        )
        order_xml = eTree.XML(order_picture)
        order = order_xml.find("Order") if order_xml.find("Order") else order_xml

    if not has_electronic_payment(order):
        return False

    if operator_id is None:
        operator = model.find("Operator")
        operator_id = operator.get("id", "") if operator is not None else ""

    if tender is not None:
        return _cancel_eft_tender(
            pos_id,
            order,
            tender,
            is_table_service,
            is_mobile,
            operator_id
        )

    results = []
    tenders = get_electronic_tenders(order)
    order_interactor = OrderInteractor(pos_id)
    for tender in tenders:
        tender_amount = float(tender.get("amount"))
        if tender_amount == 0:
            continue

        result = _cancel_eft_tender(
            pos_id,
            order,
            tender,
            is_table_service,
            is_mobile,
            operator_id
        )
        results.append(result)
        order_interactor.clear_order_tender(
            order_tender_id=tender.get("id"),
            order_id=order_id or order.get("orderId"),
        )

    return all(results)


def _cancel_eft_tender(
    pos_id,             # type: str
    order,              # type: eTree.XML
    tender,             # type: eTree.XML
    is_table_service,   # type: bool
    is_mobile,          # type: bool
    operator_id         # type: str
):
    # type: (...) -> bool

    data, tender_id = _get_cancel_eft_tender_data(
        pos_id,
        order,
        tender,
        is_table_service,
        operator_id
    )
    try:
        component_name = get_electronic_payment_component_name(pos_id=pos_id)
        if component_name is None:
            return False

        timeout = 600000000
        token = TK_ELECTRONIC_PAYMENT_CANCEL_PAYMENT_ORDER
        result = mbcontext.MB_EasySendMessage(component_name, token, FM_PARAM, data, timeout)
        if result.token == TK_SYS_ACK:
            if is_mobile:
                return True

            success_update_tenders = save_cancel_receipt(
                pos_id,
                order,
                tender_id,
                result.data
            )
            if success_update_tenders:
                return True

        logger.error("Error cancel payment: {} - Tender: {}".format(result.data, tender_id))
    except Exception as _:
        logger.exception("Error cancel payment: Tender: {}".format(tender_id))
        show_messagebox(pos_id, "$ERROR_CANCEL_ELECTRONIC_PAYMENTS_ORDER")

    return False


def _get_cancel_eft_tender_data(
    pos_id,             # type: str
    order,              # type: eTree.XML
    tender,             # type: eTree.XML
    is_table_service,   # type: bool
    operator_id         # type: str
):
    # type: (...) -> Tuple[str, str]

    default_tender_id = tender.get("tenderId") or tender.get("id", "")
    tender_id = default_tender_id if is_table_service else ""
    tender_type = tender.get("type") or tender.get("tenderType")
    tender_amount = float(tender.get("amount") or tender.get("tenderAmount"))
    detail = tender.get("detail") or json.loads(tender.get("tenderDetail", "{}"))
    last_totaled = order.findall("StateHistory/State[@state='TOTALED']")[-1]
    last_totaled_date = last_totaled.get("timestamp").replace("-", "").replace(":", "")
    electronic_type = int(tender.get("electronicType") or tender.get("eletronicType"))

    data = json.dumps(
        {
            "posId": pos_id,
            "orderId": order.get("orderId"),
            "tenderId": tender_id,
            "tenderType": tender_type,
            "tenderAmount": tender_amount,
            "fiscalDate": last_totaled_date[:8],
            "fiscalTime": last_totaled_date[9:15],
            "tenderDetail": detail,
            "operatorId": operator_id or "",
            "electronicType": electronic_type
        }
    )

    return data, default_tender_id



import codecs
import os
import sys

from ._compat import PY2


def _find_unicode_literals_frame():
    import __future__

    if not hasattr(sys, "_getframe"):  # not all Python implementations have it
        return 0
    frm = sys._getframe(1)
    idx = 1
    while frm is not None:
        if frm.f_globals.get("__name__", "").startswith("click."):
            frm = frm.f_back
            idx += 1
        elif frm.f_code.co_flags & __future__.unicode_literals.compiler_flag:
            return idx
        else:
            break
    return 0


def _check_for_unicode_literals():
    if not __debug__:
        return

    from . import disable_unicode_literals_warning

    if not PY2 or disable_unicode_literals_warning:
        return
    bad_frame = _find_unicode_literals_frame()
    if bad_frame <= 0:
        return
    from warnings import warn

    warn(
        Warning(
            "Click detected the use of the unicode_literals __future__"
            " import. This is heavily discouraged because it can"
            " introduce subtle bugs in your code. You should instead"
            ' use explicit u"" literals for your unicode strings. For'
            " more information see"
            " https://click.palletsprojects.com/python3/"
        ),
        stacklevel=bad_frame,
    )


def _verify_python3_env():
    """Ensures that the environment is good for unicode on Python 3."""
    if PY2:
        return
    try:
        import locale

        fs_enc = codecs.lookup(locale.getpreferredencoding()).name
    except Exception:
        fs_enc = "ascii"
    if fs_enc != "ascii":
        return

    extra = ""
    if os.name == "posix":
        import subprocess

        try:
            rv = subprocess.Popen(
                ["locale", "-a"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
            ).communicate()[0]
        except OSError:
            rv = b""
        good_locales = set()
        has_c_utf8 = False

        # Make sure we're operating on text here.
        if isinstance(rv, bytes):
            rv = rv.decode("ascii", "replace")

        for line in rv.splitlines():
            locale = line.strip()
            if locale.lower().endswith((".utf-8", ".utf8")):
                good_locales.add(locale)
                if locale.lower() in ("c.utf8", "c.utf-8"):
                    has_c_utf8 = True

        extra += "\n\n"
        if not good_locales:
            extra += (
                "Additional information: on this system no suitable"
                " UTF-8 locales were discovered. This most likely"
                " requires resolving by reconfiguring the locale"
                " system."
            )
        elif has_c_utf8:
            extra += (
                "This system supports the C.UTF-8 locale which is"
                " recommended. You might be able to resolve your issue"
                " by exporting the following environment variables:\n\n"
                "    export LC_ALL=C.UTF-8\n"
                "    export LANG=C.UTF-8"
            )
        else:
            extra += (
                "This system lists a couple of UTF-8 supporting locales"
                " that you can pick from. The following suitable"
                " locales were discovered: {}".format(", ".join(sorted(good_locales)))
            )

        bad_locale = None
        for locale in os.environ.get("LC_ALL"), os.environ.get("LANG"):
            if locale and locale.lower().endswith((".utf-8", ".utf8")):
                bad_locale = locale
            if locale is not None:
                break
        if bad_locale is not None:
            extra += (
                "\n\nClick discovered that you exported a UTF-8 locale"
                " but the locale system could not pick up from it"
                " because it does not exist. The exported locale is"
                " '{}' but it is not supported".format(bad_locale)
            )

    raise RuntimeError(
        "Click will abort further execution because Python 3 was"
        " configured to use ASCII as encoding for the environment."
        " Consult https://click.palletsprojects.com/python3/ for"
        " mitigation steps.{}".format(extra)
    )



from logging import Logger

from interactor.route.menu.get_banners.dto import (
    GetBannersRequest,
    GetBannersResponse
)
from model.config import Config
from validation_util import validate_api_key


class GetBannersInteractor(object):

    def __init__(
            self,
            logger,   # type: Logger
            config    # type: Config
    ):
        # type: (...) -> None

        self.logger = logger
        self.config = config

    def execute(
            self,
            request   # type: GetBannersRequest
    ):
        # type: (...) -> GetBannersResponse

        validate_api_key(request, self.config.api_key_mapping)

        return GetBannersResponse(self.config.totem_config_path)



class OrderValidationException(Exception):
    pass


class BenefitExpiredOrInactiveException(OrderValidationException):
    pass


class PodTypeNotValidException(OrderValidationException):
    pass


class InsufficientSaleAmountException(OrderValidationException):
    pass


class ErrorApplyingCoupon(OrderValidationException):
    pass


class ExcessiveSaleAmountException(OrderValidationException):
    pass


class UserLevelNotAllowedException(OrderValidationException):
    pass


class MissingItemsException(OrderValidationException):
    pass


class CustomBenefitAlreadyAppliedException(OrderValidationException):
    pass


class InvalidServiceTypeException(OrderValidationException):
    pass


class InvalidSaleTypeException(OrderValidationException):
    pass


class InvalidMinSeatsException(OrderValidationException):
    pass


class InvalidSectorException(OrderValidationException):
    pass


class ApplyingDiscountUnknownException(OrderValidationException):
    pass


class OrderPictureException(OrderValidationException):
    pass


class InvalidAddItemsInTotaledOrderException(OrderValidationException):
    pass


class InvalidServicePeriodOrderException(OrderValidationException):
    pass


class InvalidBrandException(OrderValidationException):
    pass


class AddItemsException(OrderValidationException):
    pass



from enum import Enum


class CashReportBodyLineDto(object):
    def __init__(self, quantity=0, value=0.0, detail=None):
        # type: (int, float, str) -> None
        self.quantity = quantity
        self.value = value
        self.detail = detail



# encoding: utf-8
from old_helper import BaseRepository


class ProductRepository(BaseRepository):

    def __init__(self, mbcontext):
        super(ProductRepository, self).__init__(mbcontext)

    def get_tenders_descripton(self):
        def inner_func(conn):
            query = self._TenderDescriptionQuery
            tenders = {}
            for x in conn.select(query):
                tenders[x.get_entry("TenderId")] = x.get_entry("TenderDescr")
            return tenders

        return self.execute_with_connection(inner_func)

    _TenderDescriptionQuery = """\
    SELECT t1.TenderId, t1.TenderDescr || ' ' || coalesce(t2.TenderDescr, '') AS TenderDescr
    FROM TenderType AS t1
    LEFt JOIN TenderType AS t2 ON t1.ParentTenderId = t2.TenderId
    """



from __future__ import absolute_import
from future.standard_library import suspend_hooks

from future.utils import PY3

if PY3:
    from urllib.parse import *
else:
    __future_module__ = True
    from urlparse import (ParseResult, SplitResult, parse_qs, parse_qsl,
                          urldefrag, urljoin, urlparse, urlsplit,
                          urlunparse, urlunsplit)
    
    # we use this method to get at the original py2 urllib before any renaming
    # quote = sys.py2_modules['urllib'].quote
    # quote_plus = sys.py2_modules['urllib'].quote_plus
    # unquote = sys.py2_modules['urllib'].unquote
    # unquote_plus = sys.py2_modules['urllib'].unquote_plus
    # urlencode = sys.py2_modules['urllib'].urlencode
    # splitquery = sys.py2_modules['urllib'].splitquery
    
    with suspend_hooks():
        from urllib import (quote,
                            quote_plus,
                            unquote,
                            unquote_plus,
                            urlencode,
                            splitquery)



from abc import ABCMeta, abstractmethod

from typing import List


class ReportBuilder(object):

    @abstractmethod
    def generate(self):
        # type: () -> List
        raise NotImplementedError()


from mw_helper import ensure_iterable
from production.box._OrderChangerProductionBox import OrderChangerProductionBox
from production.model import ProdStates


class SaleTypeFilterBox(OrderChangerProductionBox):
    def __init__(self, name, sons, sale_types, logger=None):
        super(SaleTypeFilterBox, self).__init__(name, sons, logger)
        sale_types = ensure_iterable(sale_types)

        self.allowed_sale_types = {}
        for sale_type in sale_types:
            self.allowed_sale_types[sale_type] = sale_type

    def change_order(self, order):
        if order.sale_type not in self.allowed_sale_types:
            order.prod_state = ProdStates.INVALID

        return order



from json import dumps
from typing import Dict, Tuple

from unidecode import unidecode

from ._Action import Action
from ._Conditions import Conditions
from ._Exclusions import Exclusions


class Benefit(object):
    def __init__(self, raw_benefit):
        # type: (Dict) -> None

        self.id = raw_benefit.get("id")
        self.name = unidecode(raw_benefit.get("name"))
        self.user_description = raw_benefit.get("userDescription")
        self.priority = int(raw_benefit.get("priority")) if raw_benefit.get("priority") else None
        self.automatic = raw_benefit.get("automatic")
        self.tip_included = raw_benefit.get("tipIncluded") if raw_benefit.get("tipIncluded") else False
        self.ask_value = bool(raw_benefit.get("askValue"))
        conditions = raw_benefit.get("conditions") if raw_benefit.get("conditions") else {}
        self.conditions = Conditions(conditions)
        actions = [self._get_action(e) for e in raw_benefit.get("actions", {}).items() if self._validate_action(e)]
        self.actions = actions
        exclusions = raw_benefit.get("exclusions") if raw_benefit.get("exclusions") else {}
        self.exclusions = Exclusions(exclusions)
        self.raw = raw_benefit
        self.times_to_apply = raw_benefit.get("timesToApply")
        self.group = raw_benefit.get("group")

    def to_json(self):
        # type: () -> str

        return dumps(self, default=lambda o: o.__dict__ if hasattr(o, '__dict__') else str(o), sort_keys=True)

    def _get_action(
        self,
        e,      # type: Tuple
    ):
        # type: (...) -> Action

        return Action(
            {
                "order": e[0],
                "name": e[1].items()[0][0],
                "value": e[1].items()[0][1],
                "seat": self._get_benefit_seat(e[1].items()),
                "maxValue": e[1]["maxValue"] if "maxValue" in e[1] else None,
            }
        )

    @staticmethod
    def _validate_action(e):
        # type: (Tuple) -> bool

        action_name = e[1].items()[0][0]
        action_value = e[1].items()[0][1]

        if action_name == 'discountAmountOnTotal':
            if action_value == 0:
                return False

        if action_name == 'discountPercentageOnTotal':
            if action_value == 0 or action_value > 100:
                return False

        if action_name == 'itemsToAdd':
            if action_value is None:
                return False

        return True

    @staticmethod
    def _get_benefit_seat(
        benefit_tuple,  # type: Tuple
    ):
        # type: (...) -> str

        seat = None

        if len(benefit_tuple) >= 2 and benefit_tuple[1][0] == "seat":
            seat = benefit_tuple[1][1]
        elif type(benefit_tuple[0][1]) == "list":
            seat = benefit_tuple[0][1][0].get("seat", None)

        return seat



from ._GetOrderMaxPrepTimeInteractor import GetOrderMaxPrepTimeInteractor
from ._GetPrintersInfoInteractor import GetPrintersInfoInteractor
from ._GetCoursesInteractor import GetCoursesInteractor
from ._PickListOrdersInteractor import PickListOrdersInteractor



from typing import List, Dict


class BenefitList(object):
    def __init__(
            self,
            benefit_pre_total_sequence,         # type: List
            manual_benefits_on_total_sequence,  # type: List
            discount_item_values,               # type: Dict
            discount_id_tip_excluded,           # type: List
            tender_id_tip_excluded,             # type: List
            benefits                            # type: List
    ):
        # type: (...) -> None

        self.benefit_pre_total_sequence = benefit_pre_total_sequence
        self.manual_benefits_on_total_sequence = manual_benefits_on_total_sequence
        self.discount_item_values = discount_item_values
        self.discount_id_tip_excluded = discount_id_tip_excluded
        self.tender_id_tip_excluded = tender_id_tip_excluded
        self.benefit_list = benefits



from xml.etree import cElementTree as eTree

import sysactions
from msgbus import MBEasyContext
from old_helper import BaseRepository

USERS_CACHE = {}


class UsersRepository(BaseRepository):

    def __init__(self, mbcontext):
        # type: (MBEasyContext) -> None
        super(UsersRepository, self).__init__(mbcontext)

    @staticmethod
    def get_operators_name(voided_lines):
        for line in voided_lines:
            if line.operator in USERS_CACHE:
                line.operator_name = USERS_CACHE[line.operator]
                continue
            user_informations = sysactions.get_user_information(line.operator)
            if user_informations:
                user_name = eTree.XML(user_informations).find(".//user").get("UserName")
                line.operator_name = USERS_CACHE[line.operator] = user_name

        return voided_lines



from datetime import datetime

from typing import List, Dict

from ._TableOrder import TableOrder


class Table(object):
    def __init__(self):
        self.table_id = None  # type: str
        self.table_name = None  # type: str
        self.start_date = None  # type: datetime
        self.service_seats = 0  # type: int
        self.total_amount = 0.0  # type: float
        self.sub_total_amount = 0.0  # type: float
        self.total_per_seat = 0.0  # type: float
        self.total_discounts = 0.0  # type: float
        self.detailed_discounts = {}  # type: Dict
        self.detailed_payments = {}  # type: Dict
        self.orders = []  # type: List[TableOrder]
        self.tip = 0.0  # type: float
        self.user_id = None  # type: str
        self.user_name = None  # type: str
        self.promotional_tenders = 0  # type: float
        self.detailed_tip_excluded = {}  # type: Dict
        self.service_tenders = []  # type: List[dict]
        self.discount_groups = {}  # type: Dict



from _LogoutSummaryHeaderDto import LogoutSummaryHeaderDto  # noqa
from _LogoutSummaryBodyDto import LogoutSummaryBodyDto  # noqa


class LogoutSummaryDto(object):
    def __init__(self, logout_summary_header_dto, logout_summary_body_dto):
        # type:(LogoutSummaryHeaderDto, LogoutSummaryBodyDto) -> ()
        self.logout_summary_header_dto = logout_summary_header_dto
        self.logout_summary_body_dto = logout_summary_body_dto



from production.command import ChangeProdStateCommand
from production.model import ProductionOrder
from production.view import ProductionView


class CommandView(ProductionView):
    def __init__(self, name, command_processor, command, logger):
        super(CommandView, self).__init__(name, logger)

        self.command_processor = command_processor
        self.command = command

    def handle_order(self, order, synch=False):
        # type: (ProductionOrder, bool) -> None

        production_command = ChangeProdStateCommand(order.order_id, None, self.command)
        self.command_processor.handle_command(production_command)

    def get_view_tags(self):
        return []



import os
import json


class GetBannersResponse(object):
    def __init__(
        self,
        totem_config_path  # type: str
    ):
        # type: (...) -> None

        if not totem_config_path or not os.path.exists(totem_config_path):
            return

        config_file = open(totem_config_path, 'rb')
        data = json.load(config_file)
        self.banners = data.get("banner", {})



from datetime import datetime

from typing import List
from typing import Union

from _OrderTender import OrderTender


class OrderSate(object):
    VOIDED = 4
    PAID = 5


class Order(object):
    def __init__(self, order_id, date, state, total, donation, tenders, void_reason_id, discount, tip, sale_type, table_id, sector, delivery_originator=None, nf_number=None, fiscal_canceled=False):
        # type: (int, datetime, int, float, float, List[OrderTender], Union[int, None], float, float, int, str, Union[str, None], Union[str, None], int, bool) -> None
        self.order_id = order_id
        self.date = date
        self.state = state
        self.total = total
        self.donation = donation
        self.tenders = tenders
        self.void_reason_id = void_reason_id
        self.discount = discount
        self.tip = tip
        self.sale_type = sale_type
        self.table_id = table_id
        self.sector = sector
        self.delivery_originator = delivery_originator
        self.nf_number = nf_number
        self.fiscal_canceled = fiscal_canceled



from datetime import datetime


class OpenedTable(object):
    def __init__(
        self,
        table_id,       # type: str
        name,           # type: str
        sector,         # type: str
        opened_date,    # type: datetime
        status,         # type: str
        operator,       # type: str
        seats,          # type: seats
    ):
        # type: (...) -> None

        self.table_id = table_id
        self.name = name
        self.sector = sector
        self.opened_date = opened_date
        self.status = status
        self.operator = operator
        self.amount = 0
        self.seats = seats
        self.orders = []



from ._I18n import I18n
from ._SimpleReport import SimpleReport
from ._TableReport import TableReport
from ._I18nReporter import I18nReporter
from ._ReportColumnDefinition import ReportColumnDefinition
from ._AlignTypes import AlignTypes



import math

from abc import ABCMeta
from typing import List, Any

from ._OrderChangerProductionBox import OrderChangerProductionBox
from production.model import ProductionOrder, Item


class OrderSplitterBox(OrderChangerProductionBox):
    __metaclass__ = ABCMeta

    def __init__(self, name, sons, order_changer, split_type, split_value, logger=None):
        super(OrderSplitterBox, self).__init__(name, sons, logger)
        self.order_changer = order_changer
        self.split_type = split_type
        self.split_value = split_value

    def change_order(self, order):
        # type: (ProductionOrder) -> List[ProductionOrder]

        if not self._possible_to_split_order(order) or self.split_type is None:
            return [order]

        if self.split_type.upper() == 'PREP_TIME':
            orders = self.split_prep_time(order)
        elif self.split_type.upper() == 'SALE_LINES':
            if self.split_value is None:
                return [order]
            orders = self.split_sale_line(order, float(self.split_value))

        elif self.split_type.upper() in 'FIRE_HOLD_PRIORITY':
            orders = self.split_fire_hold_priority(order)
        else:
            orders = [order]

        return orders

    @staticmethod
    def split_prep_time(order):
        # type: (ProductionOrder) -> List[ProductionOrder]

        orders = []
        split_index = 0

        for item in order.items:
            for temp_order in orders:
                if len(temp_order.items) > 0 and item.prep_time == temp_order.items[0].prep_time:
                    temp_order.items.append(item)
                    break
            else:
                new_order = order.clone([item], split_index)
                orders.append(new_order)
                split_index += 1

        return orders

    def split_sale_line(self, order, limit):
        # type: (ProductionOrder, float) -> List[ProductionOrder]

        temp_items = [x for x in order.items if not self._has_fire_hold_or_priority(x)]
        if len(temp_items) < limit:
            return [order]

        orders = []
        split_index = 0
        box_limit = math.ceil(len(temp_items) / limit)
        items_per_box = math.ceil(len(temp_items) / box_limit)
        for item in temp_items:
            if len(orders) <= split_index:
                new_order = order.clone([], split_index)
                orders.append(new_order)

            orders[split_index].items.append(item)
            if len(orders[split_index].items) >= items_per_box:
                split_index += 1

        orders[0].items.extend([x for x in order.items if self._has_fire_hold_or_priority(x)])
        return orders

    def split_fire_hold_priority(self, order):
        # type: (ProductionOrder) -> List[ProductionOrder]

        if not self._possible_to_split_order(order):
            return [order]

        orders = []
        main_order_items = []
        split_index = 1
        for item in order.items:
            if not self._has_fire_hold_or_priority(item):
                main_order_items.append(item)
                continue

            hold_item = self._get_item_with_property(item.clone(), self._hold_in_item)
            if hold_item:
                new_hold_order = order.clone([hold_item], split_index)
                clear_tag_time = item.get_last_tag_time("clear_hold")
                if clear_tag_time is not None:
                    new_hold_order.display_time = clear_tag_time.strftime("%Y-%m-%dT%H:%M:%S")
                orders.append(new_hold_order)
                split_index += 1

            fire_item = self._get_item_with_property(item.clone(), self._fire_in_item)
            if fire_item:
                new_fire_order = order.clone([fire_item], split_index)
                orders.append(new_fire_order)
                split_index += 1

            priority_item = self._get_item_with_property(item.clone(), self._priority_in_item)
            if priority_item:
                new_priority_order = order.clone([priority_item], split_index)
                orders.append(new_priority_order)
                split_index += 1

            if hold_item or fire_item or priority_item:
                property_list = [self._hold_in_item, self._fire_in_item, self._priority_in_item]
                clear_item = self._get_item_without_properties(item.clone(), property_list)
                if clear_item:
                    main_order_items.append(clear_item)

        self.order_changer.change_order(order.order_id, self._mark_items_as_split)

        new_order = order
        if len(orders) > 0:
            new_order = order.clone()
            new_order.order_id = str(order.order_id) + "-0"
        new_order.items = main_order_items
        orders.append(new_order)

        return orders

    def _mark_items_as_split(self, root):
        # type: (Any[ProductionOrder, Item]) -> bool

        need_to_save_order = False
        child_need_to_save_order = False
        for item in root.items:
            need_to_save_order = self._mark_items_as_split(item)

            if "fire" in item.properties and "had_fire" not in item.tags:
                item.add_tag("had_fire")
                need_to_save_order = True

            if "hold" in item.properties and "had_hold" not in item.tags:
                item.add_tag("had_hold")
                need_to_save_order = True

            if "hold" not in item.properties and "had_hold" in item.tags and "clear_hold" not in item.tags:
                item.add_tag("clear_hold")
                need_to_save_order = True

            if "highPriority" in item.properties and "had_high_priority" not in item.tags:
                item.add_tag("had_high_priority")
                need_to_save_order = True

        return need_to_save_order or child_need_to_save_order

    @staticmethod
    def _possible_to_split_order(order):
        lines = 0
        has_combo = False
        for item in order.items:
            lines += 1 if item.qty > 0 else 0
            has_combo = True if item.item_type == 'COMBO' else has_combo
        return lines > 1 or has_combo

    @staticmethod
    def _hold_in_item(item):
        return "hold" in item.properties or "had_hold" in item.tags

    @staticmethod
    def _fire_in_item(item):
        return "fire" in item.properties or "had_fire" in item.tags

    @staticmethod
    def _priority_in_item(item):
        return "highPriority" in item.properties or "had_high_priority" in item.tags

    def _has_fire_hold_or_priority(self, item):
        has_fire = self._fire_in_item(item)
        has_hold = self._hold_in_item(item)
        has_high_priority = self._priority_in_item(item)

        children_has_split = False
        for child in item.items:
            children_has_split = children_has_split or self._has_fire_hold_or_priority(child)
            if children_has_split:
                break

        return has_fire or has_hold or has_high_priority or children_has_split

    def _get_item_with_property(self, item, item_has_property):

        if item_has_property(item):
            return item

        if not item.items:
            return None

        children = []
        for child in item.items:
            tag_child = self._get_item_with_property(child, item_has_property)
            if tag_child:
                children.append(tag_child)
        item.items = children
        return item if children else None

    def _get_item_without_properties(self, item, item_has_properties):
        if any([f(item) for f in item_has_properties]):
            return None

        if not item.items:
            return item

        children = []
        for child in item.items:
            tag_child = self._get_item_without_properties(child, item_has_properties)
            if tag_child:
                children.append(tag_child)
        item.items = children
        return item if children else None



from deliveryreport.model import DeliveryOrder


class DeliveryBodyDto(object):
    def __init__(
        self,
        order,  # type: DeliveryOrder
    ):
        # type: (...) -> None

        self.items = order.items
        self.total_price = order.total_price
        self.sub_total_price = order.sub_total_price
        self.delivery_fee = order.delivery_fee
        self.additional_fee = order.additional_fee
        self.observation = order.observation
        self.observations = order.observations
        self.tenders = order.tenders
        self.customer_name = order.customer_name
        self.customer_doc = order.customer_doc
        self.customer_phone = order.customer_phone
        self.address = order.address
        self.address_neighborhood = order.address_neighborhood
        self.city = order.city
        self.address_complement = order.address_complement
        self.postal_code = order.postal_code
        self.vouchers = order.vouchers
        self.partner = order.partner
        self.need_logistics = order.need_logistics



from ._GetLoyaltyIdInteractor import GetLoyaltyIdInteractor
from ._GetAndLockVoucherInteractor import GetAndLockVoucherInteractor
from ._BurnVoucherInteractor import BurnVoucherInteractor
from ._UnlockVoucherInteractor import UnlockVoucherInteractor
from ._LoyaltyAvailableInteractor import LoyaltyAvailableInteractor



class InvalidResponseMessage(Exception):
    def __init__(self, reason):
        self.reason = reason

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "InvalidResponseMessage(reason='')".format(self.reason)



# encoding: utf-8
from old_helper import BaseRepository

from openedtablesreport.model import OpenedTable


class TableRepository(BaseRepository):

    def __init__(self, mbcontext, pos_id):
        super(TableRepository, self).__init__(mbcontext)
        self.pos_id = pos_id

    def get_opened_tables(self):
        # type: (...) -> OpenedTable

        def inner_func(conn):
            query = self._OpenedTablesQuery
            tables = {}
            for x in conn.select(query):
                table_id = x.get_entry("TableId")
                table_status = x.get_entry("Status")
                split_order_id = x.get_entry("SplitOrderId")
                order_id = x.get_entry("OrderId")

                if table_status == "TOTALED":
                    if split_order_id is None:
                        continue

                    order_id = split_order_id
                else:
                    if split_order_id is not None:
                        continue

                if table_id not in tables:
                    tables[table_id] = OpenedTable(
                        table_id=table_id,
                        name=x.get_entry("TableName"),
                        sector=x.get_entry("Sector"),
                        opened_date=x.get_entry("StartTS"),
                        status=table_status,
                        operator=x.get_entry("UserId"),
                        seats=x.get_entry("NumberOfSeats"),
                    )

                if order_id not in tables[table_id].orders:
                    tables[table_id].orders.append(order_id)

            return tables

        return self.execute_with_connection(inner_func)

    _OpenedTablesQuery = """
        SELECT DISTINCT
            rt.TableId,
            COALESCE(scp.PropertyValue, rt.TableName, rt.TableId) AS TableName,
            rt.sector,
            ts.StartTS,
            UPPER(ts.Descr) AS Status,
            ts.UserId,
            ts.NumberOfSeats,
            so.OrderId,
            ss.SplitOrderId
        FROM TableService ts
        LEFT JOIN ServiceOrders so ON so.ServiceId = ts.ServiceId AND so.state != 'IN_PROGRESS'
        JOIN RestaurantTable rt ON ts.TableId = rt.TableId
        LEFT JOIN ServiceSplit ss ON ts.ServiceId = ss.ServiceId AND so.OrderId = ss.OrderId
        JOIN TableStatus ts ON rt.Status = ts.Status
        LEFT JOIN ServiceCustomProperty scp ON ts.ServiceId = scp.ServiceId AND scp.PropertyKey = 'TAB_ID'
        WHERE ts.FinishedTS IS NULL
    """



# -*- coding: utf-8 -*-
from datetime import datetime
from typing import Optional, List

import simplejson as json
from application.configurator import Configurator
from application.model import EventData, TransferType
from application.repository import MessageBusRepository
from iso8601 import parse_date


class OperatorLogoutInteractor(object):

    def __init__(self, configurations, repository):
        # type: (Configurator, MessageBusRepository) -> None

        self.configs = configurations
        self.logger = self.configs.logger
        self.repository = repository

    def build(self, event_data):
        # type: (EventData) -> Optional[str]

        session_id = event_data.session_id
        session_user = self._get_user_from_session_id(session_id)
        if not session_id or session_user == "0":
            self.logger.error("Invalid 'sessionid' in the 'OPERATORLOGOUT' event. Dismissing it")
            return

        business_period = event_data.data.get("period")
        payload_period = datetime.strptime(business_period, "%Y%m%d")
        pos_id = event_data.data.get("posid")
        operator_time = self.repository.get_operator_time(session_id)
        op_open = operator_time.get("OpenTime")
        op_close = operator_time.get("CloseTime")

        payload = {
            "store-code": str(self.configs.store_id),
            "business-dt": payload_period.strftime("%Y-%m-%d"),
            "pos-code": pos_id,
            "pos-user-id": int(event_data.data.get("operatorid")),
            "session": session_id,
            "opened-at-dttm": op_open,
            "closed-at-dttm": op_close,
            "closed-drawer-detail": self._get_closed_drawer_detail(pos_id, session_id),
            "paid-in-out-list": self._get_paid_in_out_list(event_data),
            "justification": self._get_closed_drawer_justification(
                pos_id=pos_id,
                session_id=session_id,
            )
        }

        return json.dumps(payload)

    def _get_paid_in_out_list(self, event_data):
        # type: (EventData) -> List

        paid_in_out_list = []
        session_id = event_data.data.get("sessionid")

        transfers = self.repository.get_cash_in_out_by_session_id(session_id)
        for transfer in transfers:
            data = {
                "transaction-type": transfer.get("description"),
                "transaction-amount": transfer.get("amount"),
                "transaction-dttm": parse_date(transfer.get("timestamp")).isoformat(),
                "tender-type": transfer.get("tender_type"),
                "authorized-by": int(transfer.get("manager_id")) if transfer.get("manager_id") else None
            }
            paid_in_out_list.append(data)

        return paid_in_out_list

    def _get_closed_drawer_detail(self, pos_id, session_id):
        # type: (str, str) -> List

        payload = []
        transfers_by_session_id = self.repository.get_transfers_by_session_id(pos_id, session_id)
        transfer_order_info = self.repository.get_transfer_order_info(session_id)
        for tender_id, transfer in transfers_by_session_id.iteritems():
            total_tender = 0
            tip_amount = 0
            payment_quantity = 0
            if tender_id in transfer_order_info:
                order_info = transfer_order_info[tender_id]
                total_tender = order_info.total_tender
                tip_amount = order_info.tip_amount
                payment_quantity = order_info.payment_quantity

            payload.append(
                {
                    "tender-type-code": str(tender_id),
                    "computed-amount": total_tender,
                    "tip-amount": tip_amount,
                    "transaction-count": payment_quantity,
                    "initial-float-amount": self._get_transfer_amount(transfer, TransferType.INITIAL_AMOUNT.value),
                    "paid-in-amount": self._get_transfer_amount(transfer, TransferType.TRANSFER_CASH_IN.value),
                    "paid-out-amount": self._get_transfer_amount(transfer, TransferType.TRANSFER_CASH_OUT.value),
                    "skim-amount":self._get_transfer_amount(transfer, TransferType.TRANSFER_SKIM.value),
                    "counted-amount": self._get_transfer_amount(transfer, TransferType.DECLARED_AMOUNT.value)
                }
            )

        return payload

    def _get_closed_drawer_justification(
        self,
        pos_id,         # type: str
        session_id,     # type: str
    ):
        # type: (...) -> str

        return self.repository.get_transfer_justification(
            pos_id=pos_id,
            session_id=session_id,
            transfer_description="FINAL_AMOUNT",
        )

    @staticmethod
    def _get_transfer_amount(transfer, transfer_type):
        if transfer_type in transfer:
            return transfer[transfer_type]

        return 0

    @staticmethod
    def _get_user_from_session_id(session_id):
        try:
            session_user = None
            split_session_user = session_id.split("user=")
            if len(split_session_user) > 1:
                session_user = split_session_user[1].split(",")[0]
            return session_user
        except Exception as _:
            return None



from abc import ABCMeta, abstractmethod


class Cleaner(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def clean_idle_listeners(self):
        # type: () -> None
        raise NotImplementedError()



from services.services import UpdateService


class UpdateAlertsInteractor(object):
    def __init__(
        self,
        update_service,         # type: UpdateService
    ):
        # type: (...) -> None

        self.update_service = update_service

    def process_request(
        self,
    ):
        # type: (...) -> None

        self.update_service.update_configured_alerts()



from abc import abstractmethod
from typing import Optional, Any

from messagebus import MessageBus, Message, Event
from messageprocessorutil import UuidMessageProcessor


class AsyncUuidMessageProcessor(UuidMessageProcessor):

    @abstractmethod
    def parse_data(self, data, event_type):
        # type: (str, str) -> Any
        
        pass

    @abstractmethod
    def call_business(self, obj):
        # type: (Any) -> None
        
        pass

    @abstractmethod
    def format_response(self, message_bus, message, event, input_obj, result):
        # type: (MessageBus, Optional[Message], Optional[Event], Any, Any) -> None
        
        pass

    @abstractmethod
    def format_exception(self, message_bus, message, event, input_obj, exception):
        # type: (MessageBus, Optional[Message], Optional[Event], Any, BaseException) -> None
        
        pass

    @abstractmethod
    def format_parse_exception(self, message_bus, message, event, data, exception):
        # type: (MessageBus, Optional[Message], Optional[Event], bytes, BaseException) -> None
        
        pass



from report.command import Command  # noqa
from typing import List, Optional  # noqa


class Part(object):
    def __init__(self, text=None, commands=None):
        # type: (Optional[str], Optional[List[Command]]) -> None
        self.text = text
        self.commands = commands



import os

from msgbus import MBEasyContext

from configurator import Configurator
from helper import import_pydevd
from mbcontextmessagehandler import (
    MbContextMessageBus,
    MbContextMessageHandler,
)
from messagehandlerbuilder import DeviceManagerMessageHandlerBuilder


def main():
    # type: (...) -> None

    loader_path = os.environ["LOADERCFG"]

    import_pydevd(loader_path=loader_path, port=9167)

    configurator = Configurator(loader_path=loader_path)
    configs = configurator.read_loader_configs()
    logger = configurator.get_loggers()

    mb_context = MBEasyContext(name="DeviceManager")
    message_bus = MbContextMessageBus(mbcontext=mb_context)

    message_handler_builder = DeviceManagerMessageHandlerBuilder(
        configs=configs,
        message_bus=message_bus,
        mb_context=mb_context,
        logger=logger,
        input_logger=logger,
        callback_logger=logger,
    )
    message_handler = MbContextMessageHandler(
        message_bus=message_bus,
        service_name="DeviceManager",
        service_type="DeviceManager",
        required_services="",
        event_handler_builder=message_handler_builder,
    )

    logger.info("Device Manager started...")

    return message_handler.handle_events()



from _FiscalRepository import FiscalRepository
from _OrderRepository import OrderRepository
from _TenderRepository import TenderRepository

__all__ = ["OrderRepository", "FiscalRepository", "TenderRepository"]



import json

from application.interactor import AddBenefitToApplyInteractor
from application.model import Configurations
from application.model.transactions import AddBenefitToApplyUnitOfWork
from application.service import OrderService, BenefitService
from messagebus import DefaultToken, DataType, Message
from messageprocessorutil import UuidMessageProcessor


class AddBenefitProcessor(UuidMessageProcessor):
    def __init__(self, configs, interactor, order_service, benefit_service):
        # type: (Configurations, AddBenefitToApplyInteractor, OrderService, BenefitService) -> None

        super(AddBenefitProcessor, self).__init__()
        self.configs = configs
        self.interactor = interactor
        self.order_service = order_service
        self.benefit_service = benefit_service
        self.logger = configs.logger

    def parse_data(self, data, event_type):
        # type: (str, str) -> AddBenefitToApplyUnitOfWork

        parsed_data = json.loads(data)
        pos_id = int(parsed_data.get("posId"))
        raw_benefit = parsed_data.get("benefit", {})
        benefit_applier = parsed_data.get("benefit_applier")
        tender_id = parsed_data.get("tender_id")
        user_auth_level = parsed_data.get("user_auth_level")

        return self.benefit_service.get_unit_of_work(
                pos_id,
                raw_benefit,
                benefit_applier=benefit_applier,
                tender_id=tender_id,
                user_auth_level=user_auth_level
        )

    def call_business(self, uow):
        # type: (AddBenefitToApplyUnitOfWork) -> None

        self.interactor.apply(uow)

    def format_response(self, message_bus, message, event, request, response):
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_ACK, DataType.string, ""))

    def format_exception(self, message_bus, message, event, input_obj, exception):
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))

    def format_parse_exception(self, message_bus, message, event, data, exception):
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))



from ._OrderChangerProductionBox import OrderChangerProductionBox


class OrderStateTimerBox(OrderChangerProductionBox):

    def __init__(self, name, sons, order_state, logger=None):
        super(OrderStateTimerBox, self).__init__(name, sons, logger)
        self.order_state = order_state

    def change_order(self, order):
        for order_state in order.state_history:
            if order_state.state == self.order_state:
                if order.display_time in [None, "", " "]:
                    order.display_time = order_state.timestamp
                return order

        return order



from __future__ import absolute_import, division, print_function

import sys


__all__ = ["ABC", "text_type", "PY2"]

PY2 = sys.version_info[0] == 2


if PY2:
    import abc

    text_type = unicode  # noqa

    class ABC(object):
        __metaclass__ = abc.ABCMeta


else:
    from abc import ABC

    text_type = str



from domain import StoreRetriever, Store
from old_helper import read_swconfig


class MwStoreRetriever(StoreRetriever):
    def __init__(self, mbcontext):
        self.mbcontext = mbcontext

    def get_store(self):
        code = read_swconfig(self.mbcontext, "Store.Id")
        return Store(code)



from escposformatter import DefaultEscPosFormatter
from htmlformatter import HtmlFormatter
from report import ReportParser
from sysactions import generate_report, print_text, show_print_preview, show_messagebox

from enum import Enum

from utilfunctions import (
    get_device_service,
    get_sector_name,
)


class PrinterType(Enum):
    esc_pos = "ESC_POS"
    apos = "APOS"


def print_report(
    pos_id,
    report_name,
    preview,
    order,
    *args
):
    report_data = generate_report(report_name, pos_id, *args)
    report = ReportParser().parse(report_data)

    sector_name = get_sector_name(order=order) if order else ""
    current_printer = get_device_service(
        pos_id=pos_id,
        device_type="printer",
        sector=sector_name,
    )

    if report is not None:
        if preview:
            print_preview_report = HtmlFormatter().format(report)
            if show_print_preview(pos_id, print_preview_report.encode("utf-8")) in (1, None):
                return False

        text = DefaultEscPosFormatter().format(report)
        print_text(
            posid=pos_id,
            model=None,
            text=text,
            force_printer=current_printer,
        )
    elif report is None:
        show_messagebox(pos_id, "$NO_REPORT_DATA", "$ERROR")
    else:
        return print_text(
            posid=pos_id,
            model=None,
            text=report_data,
            preview=preview,
            force_printer=current_printer,
        )



# -*- coding: utf-8 -*-
import actions.util
import sysactions
from actions.goals import update_daily_goals
from actions.login.open_operator_min_level import get_min_level_to_open_operator
from utils import (
    open_user,
    login_user,
    print_open_day_report,
    get_pod_type_and_pos_function,
    change_pod_type_and_pos_function,
    check_current_day,
    get_initial_float,
    save_initial_float,
    user_inserted_initial_amount,
    get_authorization_if_necessary, save_last_authorization, validate_user_level
)


@sysactions.action
def sign_in_table_user(pos_id, user_info):
    validate_user_level(pos_id, user_info)
    return sign_in_user(pos_id, user_info)


def sign_in_user(pos_id, user_info):
    user_id = user_info.get("UserId")
    password = user_info.get("Password")
    long_name = user_info.get("LongName")
    authorization_response = None

    if not check_current_day(pos_id):
        return False

    try:
        sysactions.authenticate_user(user_id, password)
    except sysactions.AuthenticationFailed:
        sysactions.show_messagebox(pos_id, "$INVALID_USERNAME_OR_PASSWORD", "$INFORMATION")
        return False

    user_opened = False
    pos_user = actions.util.get_opened_pos_user(pos_id, user_id)
    if pos_user is None:
        min_level = get_min_level_to_open_operator("TS")
        model = sysactions.get_model(pos_id)
        authorization_response = get_authorization_if_necessary(model, pos_id, user_info, min_level)
        if not authorization_response.success:
            return False

        confirmation_message = "$CONFIRM_OPEN_USER|{}".format(long_name)
        resp = sysactions.show_messagebox(pos_id, confirmation_message, buttons="$OK|$CANCEL", asynch=False)
        if resp is None or resp == 1:
            return False

        user_opened = open_user(pos_id, user_id)

    pod_type, pos_function = get_pod_type_and_pos_function(pos_id)
    if pod_type is None or pos_function is None:
        return False

    initial_float = None
    model = sysactions.get_model(pos_id)

    _, _, session = actions.util.get_operator_info(model, user_id)
    if pos_function != "OT" and not user_inserted_initial_amount(pos_id, session):
        initial_float = get_initial_float(pos_id, pod_type, pos_function)

    if not login_user(pos_id, user_id):
        return False

    if authorization_response:
        save_last_authorization(pos_id, authorization_response.authorized_by, authorization_response.authorized_by_level)

    change_pod_type_and_pos_function(pos_id, pod_type, pos_function)

    if initial_float is not None:
        save_initial_float(initial_float, pos_id, user_id)

    update_daily_goals(
        pos_id=pos_id,
        synchronous=False,
    )

    if user_opened:
        print_open_day_report(pos_id, initial_float)

    return True



from typing import Optional


class TableReportSaleLineDto(object):

    def __init__(self):
        self.product_name = ""              # type: str
        self.price = 0.0                    # type: float
        self.qty = 0                        # type: float
        self.default_qty = 0                # type: Optional[int]
        self.item_type = ""                 # type: str
        self.seat = None                    # type: Optional[int]
        self.level = 0                      # type: int
        self.comment = ""                   # type: Optional[str]
        self.line_number = 0                # type: int
        self.service_period = None          # type: Optional[str]
        self.order_id = None                # type: Optional[int]
        self.item_id = None                 # type: Optional[str]
        self.part_code = None               # type: Optional[int]
        self.overwritten_unit_price = None  # type: Optional[float]
        self.fraction_parent = None         # type: Optional[int]



# coding: utf-8

"""
ASN.1 type classes for PKCS#12 files. Exports the following items:

 - CertBag()
 - CrlBag()
 - Pfx()
 - SafeBag()
 - SecretBag()

Other type classes are defined that help compose the types listed above.
"""

from __future__ import unicode_literals, division, absolute_import, print_function

from .algos import DigestInfo
from .cms import ContentInfo, SignedData
from .core import (
    Any,
    BMPString,
    Integer,
    ObjectIdentifier,
    OctetString,
    ParsableOctetString,
    Sequence,
    SequenceOf,
    SetOf,
)
from .keys import PrivateKeyInfo, EncryptedPrivateKeyInfo
from .x509 import Certificate, KeyPurposeId


# The structures in this file are taken from https://tools.ietf.org/html/rfc7292

class MacData(Sequence):
    _fields = [
        ('mac', DigestInfo),
        ('mac_salt', OctetString),
        ('iterations', Integer, {'default': 1}),
    ]


class Version(Integer):
    _map = {
        3: 'v3'
    }


class AttributeType(ObjectIdentifier):
    _map = {
        # https://tools.ietf.org/html/rfc2985#page-18
        '1.2.840.113549.1.9.20': 'friendly_name',
        '1.2.840.113549.1.9.21': 'local_key_id',
        # https://support.microsoft.com/en-us/kb/287547
        '1.3.6.1.4.1.311.17.1': 'microsoft_local_machine_keyset',
        # https://github.com/frohoff/jdk8u-dev-jdk/blob/master/src/share/classes/sun/security/pkcs12/PKCS12KeyStore.java
        # this is a set of OIDs, representing key usage, the usual value is a SET of one element OID 2.5.29.37.0
        '2.16.840.1.113894.746875.1.1': 'trusted_key_usage',
    }


class SetOfAny(SetOf):
    _child_spec = Any


class SetOfBMPString(SetOf):
    _child_spec = BMPString


class SetOfOctetString(SetOf):
    _child_spec = OctetString


class SetOfKeyPurposeId(SetOf):
    _child_spec = KeyPurposeId


class Attribute(Sequence):
    _fields = [
        ('type', AttributeType),
        ('values', None),
    ]

    _oid_specs = {
        'friendly_name': SetOfBMPString,
        'local_key_id': SetOfOctetString,
        'microsoft_csp_name': SetOfBMPString,
        'trusted_key_usage': SetOfKeyPurposeId,
    }

    def _values_spec(self):
        return self._oid_specs.get(self['type'].native, SetOfAny)

    _spec_callbacks = {
        'values': _values_spec
    }


class Attributes(SetOf):
    _child_spec = Attribute


class Pfx(Sequence):
    _fields = [
        ('version', Version),
        ('auth_safe', ContentInfo),
        ('mac_data', MacData, {'optional': True})
    ]

    _authenticated_safe = None

    @property
    def authenticated_safe(self):
        if self._authenticated_safe is None:
            content = self['auth_safe']['content']
            if isinstance(content, SignedData):
                content = content['content_info']['content']
            self._authenticated_safe = AuthenticatedSafe.load(content.native)
        return self._authenticated_safe


class AuthenticatedSafe(SequenceOf):
    _child_spec = ContentInfo


class BagId(ObjectIdentifier):
    _map = {
        '1.2.840.113549.1.12.10.1.1': 'key_bag',
        '1.2.840.113549.1.12.10.1.2': 'pkcs8_shrouded_key_bag',
        '1.2.840.113549.1.12.10.1.3': 'cert_bag',
        '1.2.840.113549.1.12.10.1.4': 'crl_bag',
        '1.2.840.113549.1.12.10.1.5': 'secret_bag',
        '1.2.840.113549.1.12.10.1.6': 'safe_contents',
    }


class CertId(ObjectIdentifier):
    _map = {
        '1.2.840.113549.1.9.22.1': 'x509',
        '1.2.840.113549.1.9.22.2': 'sdsi',
    }


class CertBag(Sequence):
    _fields = [
        ('cert_id', CertId),
        ('cert_value', ParsableOctetString, {'explicit': 0}),
    ]

    _oid_pair = ('cert_id', 'cert_value')
    _oid_specs = {
        'x509': Certificate,
    }


class CrlBag(Sequence):
    _fields = [
        ('crl_id', ObjectIdentifier),
        ('crl_value', OctetString, {'explicit': 0}),
    ]


class SecretBag(Sequence):
    _fields = [
        ('secret_type_id', ObjectIdentifier),
        ('secret_value', OctetString, {'explicit': 0}),
    ]


class SafeContents(SequenceOf):
    pass


class SafeBag(Sequence):
    _fields = [
        ('bag_id', BagId),
        ('bag_value', Any, {'explicit': 0}),
        ('bag_attributes', Attributes, {'optional': True}),
    ]

    _oid_pair = ('bag_id', 'bag_value')
    _oid_specs = {
        'key_bag': PrivateKeyInfo,
        'pkcs8_shrouded_key_bag': EncryptedPrivateKeyInfo,
        'cert_bag': CertBag,
        'crl_bag': CrlBag,
        'secret_bag': SecretBag,
        'safe_contents': SafeContents
    }


SafeContents._child_spec = SafeBag



from mw_helper import ensure_list, convert_to_dict
from production.box._OrderChangerProductionBox import OrderChangerProductionBox


class ProdStateFilterBox(OrderChangerProductionBox):
    def __init__(self, name, sons, allowed_prod_states, excluded_prod_states, logger=None):
        super(ProdStateFilterBox, self).__init__(name, sons, logger)
        self.allowed_prod_states = convert_to_dict(ensure_list(allowed_prod_states))
        self.excluded_prod_states = convert_to_dict(ensure_list(excluded_prod_states))

    def change_order(self, order):
        if self.should_filter_order(order):
            order.items = []
            order.prod_state = 'INVALID'
            return order

        return order

    def should_filter_order(self, order):
        if self.allowed_prod_states and order.prod_state not in self.allowed_prod_states:
            return True
        if self.excluded_prod_states and order.prod_state in self.excluded_prod_states:
            return True
        return False



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

"""Module: barcode.itf

:Provided barcodes: Interleaved 2 of 5
"""
__docformat__ = 'restructuredtext en'

from barcode.base import Barcode
from barcode.charsets import itf
from barcode.errors import *

MIN_SIZE = 0.2
MIN_QUIET_ZONE = 6.4

class ITF(Barcode):
    """Initializes a new ITF instance.

    :parameters:
        code : String
            ITF (Interleaved 2 of 5) numeric string
        writer : barcode.writer Instance
            The writer to render the barcode (default: SVGWriter).
        narrow: Integer
            Width of the narrow elements (default: 2)
        wide: Integer
            Width of the wide elements (default: 5)
            wide/narrow must be in the range 2..3
    """

    name = 'ITF'

    def __init__(self, code, writer=None, narrow=2, wide=5):
        if not code.isdigit():
            raise IllegalCharacterError('ITF code can only contain numbers.')
        #Length must be even, prepend 0 if necessary
        if len(code) % 2 != 0:
            code = '0' + code
        self.code = code
        self.writer = writer or Barcode.default_writer()
        self.narrow = narrow
        self.wide = wide

    def __unicode__(self):
        return self.code

    __str__ = __unicode__

    def get_fullcode(self):
        return self.code

    def build(self):
        data = itf.START
        for i in range(0, len(self.code), 2):
            bars_digit = int(self.code[i])
            spaces_digit = int(self.code[i+1])
            for j in range(5):
                data += itf.CODES[bars_digit][j].upper()
                data += itf.CODES[spaces_digit][j].lower()
        data += itf.STOP
        raw = ''
        for e in data:
            if e == 'W':
                raw += '1' * self.wide
            if e == 'w':
                raw += '0' * self.wide
            if e == 'N':
                raw += '1' * self.narrow
            if e == 'n':
                raw += '0' * self.narrow
        return [raw]

    def render(self, writer_options):
        options = dict(module_width=MIN_SIZE/self.narrow, quiet_zone=MIN_QUIET_ZONE)
        options.update(writer_options or {})
        return Barcode.render(self, options)



#! python
#
# Base class and support functions used by various backends.
#
# This file is part of pySerial. https://github.com/pyserial/pyserial
# (C) 2001-2020 Chris Liechti <cliechti@gmx.net>
#
# SPDX-License-Identifier:    BSD-3-Clause

from __future__ import absolute_import

import io
import time

# ``memoryview`` was introduced in Python 2.7 and ``bytes(some_memoryview)``
# isn't returning the contents (very unfortunate). Therefore we need special
# cases and test for it. Ensure that there is a ``memoryview`` object for older
# Python versions. This is easier than making every test dependent on its
# existence.
try:
    memoryview
except (NameError, AttributeError):
    # implementation does not matter as we do not really use it.
    # it just must not inherit from something else we might care for.
    class memoryview(object):   # pylint: disable=redefined-builtin,invalid-name
        pass

try:
    unicode
except (NameError, AttributeError):
    unicode = str       # for Python 3, pylint: disable=redefined-builtin,invalid-name

try:
    basestring
except (NameError, AttributeError):
    basestring = (str,)    # for Python 3, pylint: disable=redefined-builtin,invalid-name


# "for byte in data" fails for python3 as it returns ints instead of bytes
def iterbytes(b):
    """Iterate over bytes, returning bytes instead of ints (python3)"""
    if isinstance(b, memoryview):
        b = b.tobytes()
    i = 0
    while True:
        a = b[i:i + 1]
        i += 1
        if a:
            yield a
        else:
            break


# all Python versions prior 3.x convert ``str([17])`` to '[17]' instead of '\x11'
# so a simple ``bytes(sequence)`` doesn't work for all versions
def to_bytes(seq):
    """convert a sequence to a bytes type"""
    if isinstance(seq, bytes):
        return seq
    elif isinstance(seq, bytearray):
        return bytes(seq)
    elif isinstance(seq, memoryview):
        return seq.tobytes()
    elif isinstance(seq, unicode):
        raise TypeError('unicode strings are not supported, please encode to bytes: {!r}'.format(seq))
    else:
        # handle list of integers and bytes (one or more items) for Python 2 and 3
        return bytes(bytearray(seq))


# create control bytes
XON = to_bytes([17])
XOFF = to_bytes([19])

CR = to_bytes([13])
LF = to_bytes([10])


PARITY_NONE, PARITY_EVEN, PARITY_ODD, PARITY_MARK, PARITY_SPACE = 'N', 'E', 'O', 'M', 'S'
STOPBITS_ONE, STOPBITS_ONE_POINT_FIVE, STOPBITS_TWO = (1, 1.5, 2)
FIVEBITS, SIXBITS, SEVENBITS, EIGHTBITS = (5, 6, 7, 8)

PARITY_NAMES = {
    PARITY_NONE: 'None',
    PARITY_EVEN: 'Even',
    PARITY_ODD: 'Odd',
    PARITY_MARK: 'Mark',
    PARITY_SPACE: 'Space',
}


class SerialException(IOError):
    """Base class for serial port related exceptions."""


class SerialTimeoutException(SerialException):
    """Write timeouts give an exception"""


class PortNotOpenError(SerialException):
    """Port is not open"""
    def __init__(self):
        super(PortNotOpenError, self).__init__('Attempting to use a port that is not open')


class Timeout(object):
    """\
    Abstraction for timeout operations. Using time.monotonic() if available
    or time.time() in all other cases.

    The class can also be initialized with 0 or None, in order to support
    non-blocking and fully blocking I/O operations. The attributes
    is_non_blocking and is_infinite are set accordingly.
    """
    if hasattr(time, 'monotonic'):
        # Timeout implementation with time.monotonic(). This function is only
        # supported by Python 3.3 and above. It returns a time in seconds
        # (float) just as time.time(), but is not affected by system clock
        # adjustments.
        TIME = time.monotonic
    else:
        # Timeout implementation with time.time(). This is compatible with all
        # Python versions but has issues if the clock is adjusted while the
        # timeout is running.
        TIME = time.time

    def __init__(self, duration):
        """Initialize a timeout with given duration"""
        self.is_infinite = (duration is None)
        self.is_non_blocking = (duration == 0)
        self.duration = duration
        if duration is not None:
            self.target_time = self.TIME() + duration
        else:
            self.target_time = None

    def expired(self):
        """Return a boolean, telling if the timeout has expired"""
        return self.target_time is not None and self.time_left() <= 0

    def time_left(self):
        """Return how many seconds are left until the timeout expires"""
        if self.is_non_blocking:
            return 0
        elif self.is_infinite:
            return None
        else:
            delta = self.target_time - self.TIME()
            if delta > self.duration:
                # clock jumped, recalculate
                self.target_time = self.TIME() + self.duration
                return self.duration
            else:
                return max(0, delta)

    def restart(self, duration):
        """\
        Restart a timeout, only supported if a timeout was already set up
        before.
        """
        self.duration = duration
        self.target_time = self.TIME() + duration


class SerialBase(io.RawIOBase):
    """\
    Serial port base class. Provides __init__ function and properties to
    get/set port settings.
    """

    # default values, may be overridden in subclasses that do not support all values
    BAUDRATES = (50, 75, 110, 134, 150, 200, 300, 600, 1200, 1800, 2400, 4800,
                 9600, 19200, 38400, 57600, 115200, 230400, 460800, 500000,
                 576000, 921600, 1000000, 1152000, 1500000, 2000000, 2500000,
                 3000000, 3500000, 4000000)
    BYTESIZES = (FIVEBITS, SIXBITS, SEVENBITS, EIGHTBITS)
    PARITIES = (PARITY_NONE, PARITY_EVEN, PARITY_ODD, PARITY_MARK, PARITY_SPACE)
    STOPBITS = (STOPBITS_ONE, STOPBITS_ONE_POINT_FIVE, STOPBITS_TWO)

    def __init__(self,
                 port=None,
                 baudrate=9600,
                 bytesize=EIGHTBITS,
                 parity=PARITY_NONE,
                 stopbits=STOPBITS_ONE,
                 timeout=None,
                 xonxoff=False,
                 rtscts=False,
                 write_timeout=None,
                 dsrdtr=False,
                 inter_byte_timeout=None,
                 exclusive=None,
                 **kwargs):
        """\
        Initialize comm port object. If a "port" is given, then the port will be
        opened immediately. Otherwise a Serial port object in closed state
        is returned.
        """

        self.is_open = False
        self.portstr = None
        self.name = None
        # correct values are assigned below through properties
        self._port = None
        self._baudrate = None
        self._bytesize = None
        self._parity = None
        self._stopbits = None
        self._timeout = None
        self._write_timeout = None
        self._xonxoff = None
        self._rtscts = None
        self._dsrdtr = None
        self._inter_byte_timeout = None
        self._rs485_mode = None  # disabled by default
        self._rts_state = True
        self._dtr_state = True
        self._break_state = False
        self._exclusive = None

        # assign values using get/set methods using the properties feature
        self.port = port
        self.baudrate = baudrate
        self.bytesize = bytesize
        self.parity = parity
        self.stopbits = stopbits
        self.timeout = timeout
        self.write_timeout = write_timeout
        self.xonxoff = xonxoff
        self.rtscts = rtscts
        self.dsrdtr = dsrdtr
        self.inter_byte_timeout = inter_byte_timeout
        self.exclusive = exclusive

        # watch for backward compatible kwargs
        if 'writeTimeout' in kwargs:
            self.write_timeout = kwargs.pop('writeTimeout')
        if 'interCharTimeout' in kwargs:
            self.inter_byte_timeout = kwargs.pop('interCharTimeout')
        if kwargs:
            raise ValueError('unexpected keyword arguments: {!r}'.format(kwargs))

        if port is not None:
            self.open()

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    # to be implemented by subclasses:
    # def open(self):
    # def close(self):

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    @property
    def port(self):
        """\
        Get the current port setting. The value that was passed on init or using
        setPort() is passed back.
        """
        return self._port

    @port.setter
    def port(self, port):
        """\
        Change the port.
        """
        if port is not None and not isinstance(port, basestring):
            raise ValueError('"port" must be None or a string, not {}'.format(type(port)))
        was_open = self.is_open
        if was_open:
            self.close()
        self.portstr = port
        self._port = port
        self.name = self.portstr
        if was_open:
            self.open()

    @property
    def baudrate(self):
        """Get the current baud rate setting."""
        return self._baudrate

    @baudrate.setter
    def baudrate(self, baudrate):
        """\
        Change baud rate. It raises a ValueError if the port is open and the
        baud rate is not possible. If the port is closed, then the value is
        accepted and the exception is raised when the port is opened.
        """
        try:
            b = int(baudrate)
        except TypeError:
            raise ValueError("Not a valid baudrate: {!r}".format(baudrate))
        else:
            if b < 0:
                raise ValueError("Not a valid baudrate: {!r}".format(baudrate))
            self._baudrate = b
            if self.is_open:
                self._reconfigure_port()

    @property
    def bytesize(self):
        """Get the current byte size setting."""
        return self._bytesize

    @bytesize.setter
    def bytesize(self, bytesize):
        """Change byte size."""
        if bytesize not in self.BYTESIZES:
            raise ValueError("Not a valid byte size: {!r}".format(bytesize))
        self._bytesize = bytesize
        if self.is_open:
            self._reconfigure_port()

    @property
    def exclusive(self):
        """Get the current exclusive access setting."""
        return self._exclusive

    @exclusive.setter
    def exclusive(self, exclusive):
        """Change the exclusive access setting."""
        self._exclusive = exclusive
        if self.is_open:
            self._reconfigure_port()

    @property
    def parity(self):
        """Get the current parity setting."""
        return self._parity

    @parity.setter
    def parity(self, parity):
        """Change parity setting."""
        if parity not in self.PARITIES:
            raise ValueError("Not a valid parity: {!r}".format(parity))
        self._parity = parity
        if self.is_open:
            self._reconfigure_port()

    @property
    def stopbits(self):
        """Get the current stop bits setting."""
        return self._stopbits

    @stopbits.setter
    def stopbits(self, stopbits):
        """Change stop bits size."""
        if stopbits not in self.STOPBITS:
            raise ValueError("Not a valid stop bit size: {!r}".format(stopbits))
        self._stopbits = stopbits
        if self.is_open:
            self._reconfigure_port()

    @property
    def timeout(self):
        """Get the current timeout setting."""
        return self._timeout

    @timeout.setter
    def timeout(self, timeout):
        """Change timeout setting."""
        if timeout is not None:
            try:
                timeout + 1     # test if it's a number, will throw a TypeError if not...
            except TypeError:
                raise ValueError("Not a valid timeout: {!r}".format(timeout))
            if timeout < 0:
                raise ValueError("Not a valid timeout: {!r}".format(timeout))
        self._timeout = timeout
        if self.is_open:
            self._reconfigure_port()

    @property
    def write_timeout(self):
        """Get the current timeout setting."""
        return self._write_timeout

    @write_timeout.setter
    def write_timeout(self, timeout):
        """Change timeout setting."""
        if timeout is not None:
            if timeout < 0:
                raise ValueError("Not a valid timeout: {!r}".format(timeout))
            try:
                timeout + 1     # test if it's a number, will throw a TypeError if not...
            except TypeError:
                raise ValueError("Not a valid timeout: {!r}".format(timeout))

        self._write_timeout = timeout
        if self.is_open:
            self._reconfigure_port()

    @property
    def inter_byte_timeout(self):
        """Get the current inter-character timeout setting."""
        return self._inter_byte_timeout

    @inter_byte_timeout.setter
    def inter_byte_timeout(self, ic_timeout):
        """Change inter-byte timeout setting."""
        if ic_timeout is not None:
            if ic_timeout < 0:
                raise ValueError("Not a valid timeout: {!r}".format(ic_timeout))
            try:
                ic_timeout + 1     # test if it's a number, will throw a TypeError if not...
            except TypeError:
                raise ValueError("Not a valid timeout: {!r}".format(ic_timeout))

        self._inter_byte_timeout = ic_timeout
        if self.is_open:
            self._reconfigure_port()

    @property
    def xonxoff(self):
        """Get the current XON/XOFF setting."""
        return self._xonxoff

    @xonxoff.setter
    def xonxoff(self, xonxoff):
        """Change XON/XOFF setting."""
        self._xonxoff = xonxoff
        if self.is_open:
            self._reconfigure_port()

    @property
    def rtscts(self):
        """Get the current RTS/CTS flow control setting."""
        return self._rtscts

    @rtscts.setter
    def rtscts(self, rtscts):
        """Change RTS/CTS flow control setting."""
        self._rtscts = rtscts
        if self.is_open:
            self._reconfigure_port()

    @property
    def dsrdtr(self):
        """Get the current DSR/DTR flow control setting."""
        return self._dsrdtr

    @dsrdtr.setter
    def dsrdtr(self, dsrdtr=None):
        """Change DsrDtr flow control setting."""
        if dsrdtr is None:
            # if not set, keep backwards compatibility and follow rtscts setting
            self._dsrdtr = self._rtscts
        else:
            # if defined independently, follow its value
            self._dsrdtr = dsrdtr
        if self.is_open:
            self._reconfigure_port()

    @property
    def rts(self):
        return self._rts_state

    @rts.setter
    def rts(self, value):
        self._rts_state = value
        if self.is_open:
            self._update_rts_state()

    @property
    def dtr(self):
        return self._dtr_state

    @dtr.setter
    def dtr(self, value):
        self._dtr_state = value
        if self.is_open:
            self._update_dtr_state()

    @property
    def break_condition(self):
        return self._break_state

    @break_condition.setter
    def break_condition(self, value):
        self._break_state = value
        if self.is_open:
            self._update_break_state()

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    # functions useful for RS-485 adapters

    @property
    def rs485_mode(self):
        """\
        Enable RS485 mode and apply new settings, set to None to disable.
        See serial.rs485.RS485Settings for more info about the value.
        """
        return self._rs485_mode

    @rs485_mode.setter
    def rs485_mode(self, rs485_settings):
        self._rs485_mode = rs485_settings
        if self.is_open:
            self._reconfigure_port()

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    _SAVED_SETTINGS = ('baudrate', 'bytesize', 'parity', 'stopbits', 'xonxoff',
                       'dsrdtr', 'rtscts', 'timeout', 'write_timeout',
                       'inter_byte_timeout')

    def get_settings(self):
        """\
        Get current port settings as a dictionary. For use with
        apply_settings().
        """
        return dict([(key, getattr(self, '_' + key)) for key in self._SAVED_SETTINGS])

    def apply_settings(self, d):
        """\
        Apply stored settings from a dictionary returned from
        get_settings(). It's allowed to delete keys from the dictionary. These
        values will simply left unchanged.
        """
        for key in self._SAVED_SETTINGS:
            if key in d and d[key] != getattr(self, '_' + key):   # check against internal "_" value
                setattr(self, key, d[key])          # set non "_" value to use properties write function

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    def __repr__(self):
        """String representation of the current port settings and its state."""
        return '{name}<id=0x{id:x}, open={p.is_open}>(port={p.portstr!r}, ' \
               'baudrate={p.baudrate!r}, bytesize={p.bytesize!r}, parity={p.parity!r}, ' \
               'stopbits={p.stopbits!r}, timeout={p.timeout!r}, xonxoff={p.xonxoff!r}, ' \
               'rtscts={p.rtscts!r}, dsrdtr={p.dsrdtr!r})'.format(
                   name=self.__class__.__name__, id=id(self), p=self)

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    # compatibility with io library
    # pylint: disable=invalid-name,missing-docstring

    def readable(self):
        return True

    def writable(self):
        return True

    def seekable(self):
        return False

    def readinto(self, b):
        data = self.read(len(b))
        n = len(data)
        try:
            b[:n] = data
        except TypeError as err:
            import array
            if not isinstance(b, array.array):
                raise err
            b[:n] = array.array('b', data)
        return n

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    # context manager

    def __enter__(self):
        if self._port is not None and not self.is_open:
            self.open()
        return self

    def __exit__(self, *args, **kwargs):
        self.close()

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    def send_break(self, duration=0.25):
        """\
        Send break condition. Timed, returns to idle state after given
        duration.
        """
        if not self.is_open:
            raise PortNotOpenError()
        self.break_condition = True
        time.sleep(duration)
        self.break_condition = False

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    # backwards compatibility / deprecated functions

    def flushInput(self):
        self.reset_input_buffer()

    def flushOutput(self):
        self.reset_output_buffer()

    def inWaiting(self):
        return self.in_waiting

    def sendBreak(self, duration=0.25):
        self.send_break(duration)

    def setRTS(self, value=1):
        self.rts = value

    def setDTR(self, value=1):
        self.dtr = value

    def getCTS(self):
        return self.cts

    def getDSR(self):
        return self.dsr

    def getRI(self):
        return self.ri

    def getCD(self):
        return self.cd

    def setPort(self, port):
        self.port = port

    @property
    def writeTimeout(self):
        return self.write_timeout

    @writeTimeout.setter
    def writeTimeout(self, timeout):
        self.write_timeout = timeout

    @property
    def interCharTimeout(self):
        return self.inter_byte_timeout

    @interCharTimeout.setter
    def interCharTimeout(self, interCharTimeout):
        self.inter_byte_timeout = interCharTimeout

    def getSettingsDict(self):
        return self.get_settings()

    def applySettingsDict(self, d):
        self.apply_settings(d)

    def isOpen(self):
        return self.is_open

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    # additional functionality

    def read_all(self):
        """\
        Read all bytes currently available in the buffer of the OS.
        """
        return self.read(self.in_waiting)

    def read_until(self, expected=LF, size=None):
        """\
        Read until an expected sequence is found ('\n' by default), the size
        is exceeded or until timeout occurs.
        """
        lenterm = len(expected)
        line = bytearray()
        timeout = Timeout(self._timeout)
        while True:
            c = self.read(1)
            if c:
                line += c
                if line[-lenterm:] == expected:
                    break
                if size is not None and len(line) >= size:
                    break
            else:
                break
            if timeout.expired():
                break
        return bytes(line)

    def iread_until(self, *args, **kwargs):
        """\
        Read lines, implemented as generator. It will raise StopIteration on
        timeout (empty read).
        """
        while True:
            line = self.read_until(*args, **kwargs)
            if not line:
                break
            yield line


#  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
if __name__ == '__main__':
    import sys
    s = SerialBase()
    sys.stdout.write('port name:  {}\n'.format(s.name))
    sys.stdout.write('baud rates: {}\n'.format(s.BAUDRATES))
    sys.stdout.write('byte sizes: {}\n'.format(s.BYTESIZES))
    sys.stdout.write('parities:   {}\n'.format(s.PARITIES))
    sys.stdout.write('stop bits:  {}\n'.format(s.STOPBITS))
    sys.stdout.write('{}\n'.format(s))



# -*- coding: utf-8 -*-
"""
This module offers a generic date/time string parser which is able to parse
most known formats to represent a date and/or time.

This module attempts to be forgiving with regards to unlikely input formats,
returning a datetime object even for dates which are ambiguous. If an element
of a date/time stamp is omitted, the following rules are applied:
- If AM or PM is left unspecified, a 24-hour clock is assumed, however, an hour
  on a 12-hour clock (``0 <= hour <= 12``) *must* be specified if AM or PM is
  specified.
- If a time zone is omitted, a timezone-naive datetime is returned.

If any other elements are missing, they are taken from the
:class:`datetime.datetime` object passed to the parameter ``default``. If this
results in a day number exceeding the valid number of days per month, the
value falls back to the end of the month.

Additional resources about date/time string formats can be found below:

- `A summary of the international standard date and time notation
  <http://www.cl.cam.ac.uk/~mgk25/iso-time.html>`_
- `W3C Date and Time Formats <http://www.w3.org/TR/NOTE-datetime>`_
- `Time Formats (Planetary Rings Node) <http://pds-rings.seti.org/tools/time_formats.html>`_
- `CPAN ParseDate module
  <http://search.cpan.org/~muir/Time-modules-2013.0912/lib/Time/ParseDate.pm>`_
- `Java SimpleDateFormat Class
  <https://docs.oracle.com/javase/6/docs/api/java/text/SimpleDateFormat.html>`_
"""
from __future__ import unicode_literals

import datetime
import string
import time
import collections
import re
from io import StringIO
from calendar import monthrange

from six import text_type, binary_type, integer_types

from . import relativedelta
from . import tz

__all__ = ["parse", "parserinfo"]


class _timelex(object):
    # Fractional seconds are sometimes split by a comma
    _split_decimal = re.compile("([.,])")

    def __init__(self, instream):
        if isinstance(instream, binary_type):
            instream = instream.decode()

        if isinstance(instream, text_type):
            instream = StringIO(instream)

        if getattr(instream, 'read', None) is None:
            raise TypeError('Parser must be a string or character stream, not '
                            '{itype}'.format(itype=instream.__class__.__name__))

        self.instream = instream
        self.charstack = []
        self.tokenstack = []
        self.eof = False

    def get_token(self):
        """
        This function breaks the time string into lexical units (tokens), which
        can be parsed by the parser. Lexical units are demarcated by changes in
        the character set, so any continuous string of letters is considered
        one unit, any continuous string of numbers is considered one unit.

        The main complication arises from the fact that dots ('.') can be used
        both as separators (e.g. "Sep.20.2009") or decimal points (e.g.
        "4:30:21.447"). As such, it is necessary to read the full context of
        any dot-separated strings before breaking it into tokens; as such, this
        function maintains a "token stack", for when the ambiguous context
        demands that multiple tokens be parsed at once.
        """
        if self.tokenstack:
            return self.tokenstack.pop(0)

        seenletters = False
        token = None
        state = None

        while not self.eof:
            # We only realize that we've reached the end of a token when we
            # find a character that's not part of the current token - since
            # that character may be part of the next token, it's stored in the
            # charstack.
            if self.charstack:
                nextchar = self.charstack.pop(0)
            else:
                nextchar = self.instream.read(1)
                while nextchar == '\x00':
                    nextchar = self.instream.read(1)

            if not nextchar:
                self.eof = True
                break
            elif not state:
                # First character of the token - determines if we're starting
                # to parse a word, a number or something else.
                token = nextchar
                if self.isword(nextchar):
                    state = 'a'
                elif self.isnum(nextchar):
                    state = '0'
                elif self.isspace(nextchar):
                    token = ' '
                    break  # emit token
                else:
                    break  # emit token
            elif state == 'a':
                # If we've already started reading a word, we keep reading
                # letters until we find something that's not part of a word.
                seenletters = True
                if self.isword(nextchar):
                    token += nextchar
                elif nextchar == '.':
                    token += nextchar
                    state = 'a.'
                else:
                    self.charstack.append(nextchar)
                    break  # emit token
            elif state == '0':
                # If we've already started reading a number, we keep reading
                # numbers until we find something that doesn't fit.
                if self.isnum(nextchar):
                    token += nextchar
                elif nextchar == '.' or (nextchar == ',' and len(token) >= 2):
                    token += nextchar
                    state = '0.'
                else:
                    self.charstack.append(nextchar)
                    break  # emit token
            elif state == 'a.':
                # If we've seen some letters and a dot separator, continue
                # parsing, and the tokens will be broken up later.
                seenletters = True
                if nextchar == '.' or self.isword(nextchar):
                    token += nextchar
                elif self.isnum(nextchar) and token[-1] == '.':
                    token += nextchar
                    state = '0.'
                else:
                    self.charstack.append(nextchar)
                    break  # emit token
            elif state == '0.':
                # If we've seen at least one dot separator, keep going, we'll
                # break up the tokens later.
                if nextchar == '.' or self.isnum(nextchar):
                    token += nextchar
                elif self.isword(nextchar) and token[-1] == '.':
                    token += nextchar
                    state = 'a.'
                else:
                    self.charstack.append(nextchar)
                    break  # emit token

        if (state in ('a.', '0.') and (seenletters or token.count('.') > 1 or
                                       token[-1] in '.,')):
            l = self._split_decimal.split(token)
            token = l[0]
            for tok in l[1:]:
                if tok:
                    self.tokenstack.append(tok)

        if state == '0.' and token.count('.') == 0:
            token = token.replace(',', '.')

        return token

    def __iter__(self):
        return self

    def __next__(self):
        token = self.get_token()
        if token is None:
            raise StopIteration

        return token

    def next(self):
        return self.__next__()  # Python 2.x support

    @classmethod
    def split(cls, s):
        return list(cls(s))

    @classmethod
    def isword(cls, nextchar):
        """ Whether or not the next character is part of a word """
        return nextchar.isalpha()

    @classmethod
    def isnum(cls, nextchar):
        """ Whether the next character is part of a number """
        return nextchar.isdigit()

    @classmethod
    def isspace(cls, nextchar):
        """ Whether the next character is whitespace """
        return nextchar.isspace()


class _resultbase(object):

    def __init__(self):
        for attr in self.__slots__:
            setattr(self, attr, None)

    def _repr(self, classname):
        l = []
        for attr in self.__slots__:
            value = getattr(self, attr)
            if value is not None:
                l.append("%s=%s" % (attr, repr(value)))
        return "%s(%s)" % (classname, ", ".join(l))

    def __len__(self):
        return (sum(getattr(self, attr) is not None
                    for attr in self.__slots__))

    def __repr__(self):
        return self._repr(self.__class__.__name__)


class parserinfo(object):
    """
    Class which handles what inputs are accepted. Subclass this to customize
    the language and acceptable values for each parameter.

    :param dayfirst:
            Whether to interpret the first value in an ambiguous 3-integer date
            (e.g. 01/05/09) as the day (``True``) or month (``False``). If
            ``yearfirst`` is set to ``True``, this distinguishes between YDM
            and YMD. Default is ``False``.

    :param yearfirst:
            Whether to interpret the first value in an ambiguous 3-integer date
            (e.g. 01/05/09) as the year. If ``True``, the first number is taken
            to be the year, otherwise the last number is taken to be the year.
            Default is ``False``.
    """

    # m from a.m/p.m, t from ISO T separator
    JUMP = [" ", ".", ",", ";", "-", "/", "'",
            "at", "on", "and", "ad", "m", "t", "of",
            "st", "nd", "rd", "th"]

    WEEKDAYS = [("Mon", "Monday"),
                ("Tue", "Tuesday"),
                ("Wed", "Wednesday"),
                ("Thu", "Thursday"),
                ("Fri", "Friday"),
                ("Sat", "Saturday"),
                ("Sun", "Sunday")]
    MONTHS = [("Jan", "January"),
              ("Feb", "February"),
              ("Mar", "March"),
              ("Apr", "April"),
              ("May", "May"),
              ("Jun", "June"),
              ("Jul", "July"),
              ("Aug", "August"),
              ("Sep", "Sept", "September"),
              ("Oct", "October"),
              ("Nov", "November"),
              ("Dec", "December")]
    HMS = [("h", "hour", "hours"),
           ("m", "minute", "minutes"),
           ("s", "second", "seconds")]
    AMPM = [("am", "a"),
            ("pm", "p")]
    UTCZONE = ["UTC", "GMT", "Z"]
    PERTAIN = ["of"]
    TZOFFSET = {}

    def __init__(self, dayfirst=False, yearfirst=False):
        self._jump = self._convert(self.JUMP)
        self._weekdays = self._convert(self.WEEKDAYS)
        self._months = self._convert(self.MONTHS)
        self._hms = self._convert(self.HMS)
        self._ampm = self._convert(self.AMPM)
        self._utczone = self._convert(self.UTCZONE)
        self._pertain = self._convert(self.PERTAIN)

        self.dayfirst = dayfirst
        self.yearfirst = yearfirst

        self._year = time.localtime().tm_year
        self._century = self._year // 100 * 100

    def _convert(self, lst):
        dct = {}
        for i, v in enumerate(lst):
            if isinstance(v, tuple):
                for v in v:
                    dct[v.lower()] = i
            else:
                dct[v.lower()] = i
        return dct

    def jump(self, name):
        return name.lower() in self._jump

    def weekday(self, name):
        if len(name) >= min(len(n) for n in self._weekdays.keys()):
            try:
                return self._weekdays[name.lower()]
            except KeyError:
                pass
        return None

    def month(self, name):
        if len(name) >= min(len(n) for n in self._months.keys()):
            try:
                return self._months[name.lower()] + 1
            except KeyError:
                pass
        return None

    def hms(self, name):
        try:
            return self._hms[name.lower()]
        except KeyError:
            return None

    def ampm(self, name):
        try:
            return self._ampm[name.lower()]
        except KeyError:
            return None

    def pertain(self, name):
        return name.lower() in self._pertain

    def utczone(self, name):
        return name.lower() in self._utczone

    def tzoffset(self, name):
        if name in self._utczone:
            return 0

        return self.TZOFFSET.get(name)

    def convertyear(self, year, century_specified=False):
        if year < 100 and not century_specified:
            year += self._century
            if abs(year - self._year) >= 50:
                if year < self._year:
                    year += 100
                else:
                    year -= 100
        return year

    def validate(self, res):
        # move to info
        if res.year is not None:
            res.year = self.convertyear(res.year, res.century_specified)

        if res.tzoffset == 0 and not res.tzname or res.tzname == 'Z':
            res.tzname = "UTC"
            res.tzoffset = 0
        elif res.tzoffset != 0 and res.tzname and self.utczone(res.tzname):
            res.tzoffset = 0
        return True


class _ymd(list):
    def __init__(self, tzstr, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)
        self.century_specified = False
        self.tzstr = tzstr

    @staticmethod
    def token_could_be_year(token, year):
        try:
            return int(token) == year
        except ValueError:
            return False

    @staticmethod
    def find_potential_year_tokens(year, tokens):
        return [token for token in tokens if _ymd.token_could_be_year(token, year)]

    def find_probable_year_index(self, tokens):
        """
        attempt to deduce if a pre 100 year was lost
         due to padded zeros being taken off
        """
        for index, token in enumerate(self):
            potential_year_tokens = _ymd.find_potential_year_tokens(token, tokens)
            if len(potential_year_tokens) == 1 and len(potential_year_tokens[0]) > 2:
                return index

    def append(self, val):
        if hasattr(val, '__len__'):
            if val.isdigit() and len(val) > 2:
                self.century_specified = True
        elif val > 100:
            self.century_specified = True

        super(self.__class__, self).append(int(val))

    def resolve_ymd(self, mstridx, yearfirst, dayfirst):
        len_ymd = len(self)
        year, month, day = (None, None, None)

        if len_ymd > 3:
            raise ValueError("More than three YMD values")
        elif len_ymd == 1 or (mstridx != -1 and len_ymd == 2):
            # One member, or two members with a month string
            if mstridx != -1:
                month = self[mstridx]
                del self[mstridx]

            if len_ymd > 1 or mstridx == -1:
                if self[0] > 31:
                    year = self[0]
                else:
                    day = self[0]

        elif len_ymd == 2:
            # Two members with numbers
            if self[0] > 31:
                # 99-01
                year, month = self
            elif self[1] > 31:
                # 01-99
                month, year = self
            elif dayfirst and self[1] <= 12:
                # 13-01
                day, month = self
            else:
                # 01-13
                month, day = self

        elif len_ymd == 3:
            # Three members
            if mstridx == 0:
                month, day, year = self
            elif mstridx == 1:
                if self[0] > 31 or (yearfirst and self[2] <= 31):
                    # 99-Jan-01
                    year, month, day = self
                else:
                    # 01-Jan-01
                    # Give precendence to day-first, since
                    # two-digit years is usually hand-written.
                    day, month, year = self

            elif mstridx == 2:
                # WTF!?
                if self[1] > 31:
                    # 01-99-Jan
                    day, year, month = self
                else:
                    # 99-01-Jan
                    year, day, month = self

            else:
                if self[0] > 31 or \
                    self.find_probable_year_index(_timelex.split(self.tzstr)) == 0 or \
                   (yearfirst and self[1] <= 12 and self[2] <= 31):
                    # 99-01-01
                    if dayfirst and self[2] <= 12:
                        year, day, month = self
                    else:
                        year, month, day = self
                elif self[0] > 12 or (dayfirst and self[1] <= 12):
                    # 13-01-01
                    day, month, year = self
                else:
                    # 01-13-01
                    month, day, year = self

        return year, month, day


class parser(object):
    def __init__(self, info=None):
        self.info = info or parserinfo()

    def parse(self, timestr, default=None, ignoretz=False, tzinfos=None, **kwargs):
        """
        Parse the date/time string into a :class:`datetime.datetime` object.

        :param timestr:
            Any date/time string using the supported formats.

        :param default:
            The default datetime object, if this is a datetime object and not
            ``None``, elements specified in ``timestr`` replace elements in the
            default object.

        :param ignoretz:
            If set ``True``, time zones in parsed strings are ignored and a
            naive :class:`datetime.datetime` object is returned.

        :param tzinfos:
            Additional time zone names / aliases which may be present in the
            string. This argument maps time zone names (and optionally offsets
            from those time zones) to time zones. This parameter can be a
            dictionary with timezone aliases mapping time zone names to time
            zones or a function taking two parameters (``tzname`` and
            ``tzoffset``) and returning a time zone.

            The timezones to which the names are mapped can be an integer
            offset from UTC in minutes or a :class:`tzinfo` object.

            .. doctest::
               :options: +NORMALIZE_WHITESPACE

                >>> from dateutil.parser import parse
                >>> from dateutil.tz import gettz
                >>> tzinfos = {"BRST": -10800, "CST": gettz("America/Chicago")}
                >>> parse("2012-01-19 17:21:00 BRST", tzinfos=tzinfos)
                datetime.datetime(2012, 1, 19, 17, 21, tzinfo=tzoffset(u'BRST', -10800))
                >>> parse("2012-01-19 17:21:00 CST", tzinfos=tzinfos)
                datetime.datetime(2012, 1, 19, 17, 21,
                                  tzinfo=tzfile('/usr/share/zoneinfo/America/Chicago'))

            This parameter is ignored if ``ignoretz`` is set.

        :param **kwargs:
            Keyword arguments as passed to ``_parse()``.

        :return:
            Returns a :class:`datetime.datetime` object or, if the
            ``fuzzy_with_tokens`` option is ``True``, returns a tuple, the
            first element being a :class:`datetime.datetime` object, the second
            a tuple containing the fuzzy tokens.

        :raises ValueError:
            Raised for invalid or unknown string format, if the provided
            :class:`tzinfo` is not in a valid format, or if an invalid date
            would be created.

        :raises TypeError:
            Raised for non-string or character stream input.

        :raises OverflowError:
            Raised if the parsed date exceeds the largest valid C integer on
            your system.
        """

        if default is None:
            default = datetime.datetime.now().replace(hour=0, minute=0,
                                                      second=0, microsecond=0)

        res, skipped_tokens = self._parse(timestr, **kwargs)

        if res is None:
            raise ValueError("Unknown string format")

        if len(res) == 0:
            raise ValueError("String does not contain a date.")

        repl = {}
        for attr in ("year", "month", "day", "hour",
                     "minute", "second", "microsecond"):
            value = getattr(res, attr)
            if value is not None:
                repl[attr] = value

        if 'day' not in repl:
            # If the default day exceeds the last day of the month, fall back to
            # the end of the month.
            cyear = default.year if res.year is None else res.year
            cmonth = default.month if res.month is None else res.month
            cday = default.day if res.day is None else res.day

            if cday > monthrange(cyear, cmonth)[1]:
                repl['day'] = monthrange(cyear, cmonth)[1]

        ret = default.replace(**repl)

        if res.weekday is not None and not res.day:
            ret = ret+relativedelta.relativedelta(weekday=res.weekday)

        if not ignoretz:
            if (isinstance(tzinfos, collections.Callable) or
                    tzinfos and res.tzname in tzinfos):

                if isinstance(tzinfos, collections.Callable):
                    tzdata = tzinfos(res.tzname, res.tzoffset)
                else:
                    tzdata = tzinfos.get(res.tzname)

                if isinstance(tzdata, datetime.tzinfo):
                    tzinfo = tzdata
                elif isinstance(tzdata, text_type):
                    tzinfo = tz.tzstr(tzdata)
                elif isinstance(tzdata, integer_types):
                    tzinfo = tz.tzoffset(res.tzname, tzdata)
                else:
                    raise ValueError("Offset must be tzinfo subclass, "
                                     "tz string, or int offset.")
                ret = ret.replace(tzinfo=tzinfo)
            elif res.tzname and res.tzname in time.tzname:
                ret = ret.replace(tzinfo=tz.tzlocal())
            elif res.tzoffset == 0:
                ret = ret.replace(tzinfo=tz.tzutc())
            elif res.tzoffset:
                ret = ret.replace(tzinfo=tz.tzoffset(res.tzname, res.tzoffset))

        if kwargs.get('fuzzy_with_tokens', False):
            return ret, skipped_tokens
        else:
            return ret

    class _result(_resultbase):
        __slots__ = ["year", "month", "day", "weekday",
                     "hour", "minute", "second", "microsecond",
                     "tzname", "tzoffset", "ampm"]

    def _parse(self, timestr, dayfirst=None, yearfirst=None, fuzzy=False,
               fuzzy_with_tokens=False):
        """
        Private method which performs the heavy lifting of parsing, called from
        ``parse()``, which passes on its ``kwargs`` to this function.

        :param timestr:
            The string to parse.

        :param dayfirst:
            Whether to interpret the first value in an ambiguous 3-integer date
            (e.g. 01/05/09) as the day (``True``) or month (``False``). If
            ``yearfirst`` is set to ``True``, this distinguishes between YDM
            and YMD. If set to ``None``, this value is retrieved from the
            current :class:`parserinfo` object (which itself defaults to
            ``False``).

        :param yearfirst:
            Whether to interpret the first value in an ambiguous 3-integer date
            (e.g. 01/05/09) as the year. If ``True``, the first number is taken
            to be the year, otherwise the last number is taken to be the year.
            If this is set to ``None``, the value is retrieved from the current
            :class:`parserinfo` object (which itself defaults to ``False``).

        :param fuzzy:
            Whether to allow fuzzy parsing, allowing for string like "Today is
            January 1, 2047 at 8:21:00AM".

        :param fuzzy_with_tokens:
            If ``True``, ``fuzzy`` is automatically set to True, and the parser
            will return a tuple where the first element is the parsed
            :class:`datetime.datetime` datetimestamp and the second element is
            a tuple containing the portions of the string which were ignored:

            .. doctest::

                >>> from dateutil.parser import parse
                >>> parse("Today is January 1, 2047 at 8:21:00AM", fuzzy_with_tokens=True)
                (datetime.datetime(2047, 1, 1, 8, 21), (u'Today is ', u' ', u'at '))

        """
        if fuzzy_with_tokens:
            fuzzy = True

        info = self.info

        if dayfirst is None:
            dayfirst = info.dayfirst

        if yearfirst is None:
            yearfirst = info.yearfirst

        res = self._result()
        l = _timelex.split(timestr)         # Splits the timestr into tokens

        # keep up with the last token skipped so we can recombine
        # consecutively skipped tokens (-2 for when i begins at 0).
        last_skipped_token_i = -2
        skipped_tokens = list()

        try:
            # year/month/day list
            ymd = _ymd(timestr)

            # Index of the month string in ymd
            mstridx = -1

            len_l = len(l)
            i = 0
            while i < len_l:

                # Check if it's a number
                try:
                    value_repr = l[i]
                    value = float(value_repr)
                except ValueError:
                    value = None

                if value is not None:
                    # Token is a number
                    len_li = len(l[i])
                    i += 1

                    if (len(ymd) == 3 and len_li in (2, 4)
                        and res.hour is None and (i >= len_l or (l[i] != ':' and
                                                  info.hms(l[i]) is None))):
                        # 19990101T23[59]
                        s = l[i-1]
                        res.hour = int(s[:2])

                        if len_li == 4:
                            res.minute = int(s[2:])

                    elif len_li == 6 or (len_li > 6 and l[i-1].find('.') == 6):
                        # YYMMDD or HHMMSS[.ss]
                        s = l[i-1]

                        if not ymd and l[i-1].find('.') == -1:
                            #ymd.append(info.convertyear(int(s[:2])))

                            ymd.append(s[:2])
                            ymd.append(s[2:4])
                            ymd.append(s[4:])
                        else:
                            # 19990101T235959[.59]
                            res.hour = int(s[:2])
                            res.minute = int(s[2:4])
                            res.second, res.microsecond = _parsems(s[4:])

                    elif len_li in (8, 12, 14):
                        # YYYYMMDD
                        s = l[i-1]
                        ymd.append(s[:4])
                        ymd.append(s[4:6])
                        ymd.append(s[6:8])

                        if len_li > 8:
                            res.hour = int(s[8:10])
                            res.minute = int(s[10:12])

                            if len_li > 12:
                                res.second = int(s[12:])

                    elif ((i < len_l and info.hms(l[i]) is not None) or
                          (i+1 < len_l and l[i] == ' ' and
                           info.hms(l[i+1]) is not None)):

                        # HH[ ]h or MM[ ]m or SS[.ss][ ]s
                        if l[i] == ' ':
                            i += 1

                        idx = info.hms(l[i])

                        while True:
                            if idx == 0:
                                res.hour = int(value)

                                if value % 1:
                                    res.minute = int(60*(value % 1))

                            elif idx == 1:
                                res.minute = int(value)

                                if value % 1:
                                    res.second = int(60*(value % 1))

                            elif idx == 2:
                                res.second, res.microsecond = \
                                    _parsems(value_repr)

                            i += 1

                            if i >= len_l or idx == 2:
                                break

                            # 12h00
                            try:
                                value_repr = l[i]
                                value = float(value_repr)
                            except ValueError:
                                break
                            else:
                                i += 1
                                idx += 1

                                if i < len_l:
                                    newidx = info.hms(l[i])

                                    if newidx is not None:
                                        idx = newidx

                    elif (i == len_l and l[i-2] == ' ' and
                          info.hms(l[i-3]) is not None):
                        # X h MM or X m SS
                        idx = info.hms(l[i-3])

                        if idx == 0:               # h
                            res.minute = int(value)

                            sec_remainder = value % 1
                            if sec_remainder:
                                res.second = int(60 * sec_remainder)
                        elif idx == 1:             # m
                            res.second, res.microsecond = \
                                _parsems(value_repr)

                        # We don't need to advance the tokens here because the
                        # i == len_l call indicates that we're looking at all
                        # the tokens already.

                    elif i+1 < len_l and l[i] == ':':
                        # HH:MM[:SS[.ss]]
                        res.hour = int(value)
                        i += 1
                        value = float(l[i])
                        res.minute = int(value)

                        if value % 1:
                            res.second = int(60*(value % 1))

                        i += 1

                        if i < len_l and l[i] == ':':
                            res.second, res.microsecond = _parsems(l[i+1])
                            i += 2

                    elif i < len_l and l[i] in ('-', '/', '.'):
                        sep = l[i]
                        ymd.append(value_repr)
                        i += 1

                        if i < len_l and not info.jump(l[i]):
                            try:
                                # 01-01[-01]
                                ymd.append(l[i])
                            except ValueError:
                                # 01-Jan[-01]
                                value = info.month(l[i])

                                if value is not None:
                                    ymd.append(value)
                                    assert mstridx == -1
                                    mstridx = len(ymd)-1
                                else:
                                    return None, None

                            i += 1

                            if i < len_l and l[i] == sep:
                                # We have three members
                                i += 1
                                value = info.month(l[i])

                                if value is not None:
                                    ymd.append(value)
                                    mstridx = len(ymd)-1
                                    assert mstridx == -1
                                else:
                                    ymd.append(l[i])

                                i += 1
                    elif i >= len_l or info.jump(l[i]):
                        if i+1 < len_l and info.ampm(l[i+1]) is not None:
                            # 12 am
                            res.hour = int(value)

                            if res.hour < 12 and info.ampm(l[i+1]) == 1:
                                res.hour += 12
                            elif res.hour == 12 and info.ampm(l[i+1]) == 0:
                                res.hour = 0

                            i += 1
                        else:
                            # Year, month or day
                            ymd.append(value)
                        i += 1
                    elif info.ampm(l[i]) is not None:

                        # 12am
                        res.hour = int(value)

                        if res.hour < 12 and info.ampm(l[i]) == 1:
                            res.hour += 12
                        elif res.hour == 12 and info.ampm(l[i]) == 0:
                            res.hour = 0
                        i += 1

                    elif not fuzzy:
                        return None, None
                    else:
                        i += 1
                    continue

                # Check weekday
                value = info.weekday(l[i])
                if value is not None:
                    res.weekday = value
                    i += 1
                    continue

                # Check month name
                value = info.month(l[i])
                if value is not None:
                    ymd.append(value)
                    assert mstridx == -1
                    mstridx = len(ymd)-1

                    i += 1
                    if i < len_l:
                        if l[i] in ('-', '/'):
                            # Jan-01[-99]
                            sep = l[i]
                            i += 1
                            ymd.append(l[i])
                            i += 1

                            if i < len_l and l[i] == sep:
                                # Jan-01-99
                                i += 1
                                ymd.append(l[i])
                                i += 1

                        elif (i+3 < len_l and l[i] == l[i+2] == ' '
                              and info.pertain(l[i+1])):
                            # Jan of 01
                            # In this case, 01 is clearly year
                            try:
                                value = int(l[i+3])
                            except ValueError:
                                # Wrong guess
                                pass
                            else:
                                # Convert it here to become unambiguous
                                ymd.append(str(info.convertyear(value)))
                            i += 4
                    continue

                # Check am/pm
                value = info.ampm(l[i])
                if value is not None:
                    # For fuzzy parsing, 'a' or 'am' (both valid English words)
                    # may erroneously trigger the AM/PM flag. Deal with that
                    # here.
                    val_is_ampm = True

                    # If there's already an AM/PM flag, this one isn't one.
                    if fuzzy and res.ampm is not None:
                        val_is_ampm = False

                    # If AM/PM is found and hour is not, raise a ValueError
                    if res.hour is None:
                        if fuzzy:
                            val_is_ampm = False
                        else:
                            raise ValueError('No hour specified with ' +
                                             'AM or PM flag.')
                    elif not 0 <= res.hour <= 12:
                        # If AM/PM is found, it's a 12 hour clock, so raise
                        # an error for invalid range
                        if fuzzy:
                            val_is_ampm = False
                        else:
                            raise ValueError('Invalid hour specified for ' +
                                             '12-hour clock.')

                    if val_is_ampm:
                        if value == 1 and res.hour < 12:
                            res.hour += 12
                        elif value == 0 and res.hour == 12:
                            res.hour = 0

                        res.ampm = value

                    elif fuzzy:
                        last_skipped_token_i = self._skip_token(skipped_tokens,
                                                    last_skipped_token_i, i, l)
                    i += 1
                    continue

                # Check for a timezone name
                if (res.hour is not None and len(l[i]) <= 5 and
                        res.tzname is None and res.tzoffset is None and
                        not [x for x in l[i] if x not in
                             string.ascii_uppercase]):
                    res.tzname = l[i]
                    res.tzoffset = info.tzoffset(res.tzname)
                    i += 1

                    # Check for something like GMT+3, or BRST+3. Notice
                    # that it doesn't mean "I am 3 hours after GMT", but
                    # "my time +3 is GMT". If found, we reverse the
                    # logic so that timezone parsing code will get it
                    # right.
                    if i < len_l and l[i] in ('+', '-'):
                        l[i] = ('+', '-')[l[i] == '+']
                        res.tzoffset = None
                        if info.utczone(res.tzname):
                            # With something like GMT+3, the timezone
                            # is *not* GMT.
                            res.tzname = None

                    continue

                # Check for a numbered timezone
                if res.hour is not None and l[i] in ('+', '-'):
                    signal = (-1, 1)[l[i] == '+']
                    i += 1
                    len_li = len(l[i])

                    if len_li == 4:
                        # -0300
                        res.tzoffset = int(l[i][:2])*3600+int(l[i][2:])*60
                    elif i+1 < len_l and l[i+1] == ':':
                        # -03:00
                        res.tzoffset = int(l[i])*3600+int(l[i+2])*60
                        i += 2
                    elif len_li <= 2:
                        # -[0]3
                        res.tzoffset = int(l[i][:2])*3600
                    else:
                        return None, None
                    i += 1

                    res.tzoffset *= signal

                    # Look for a timezone name between parenthesis
                    if (i+3 < len_l and
                        info.jump(l[i]) and l[i+1] == '(' and l[i+3] == ')' and
                        3 <= len(l[i+2]) <= 5 and
                        not [x for x in l[i+2]
                             if x not in string.ascii_uppercase]):
                        # -0300 (BRST)
                        res.tzname = l[i+2]
                        i += 4
                    continue

                # Check jumps
                if not (info.jump(l[i]) or fuzzy):
                    return None, None

                last_skipped_token_i = self._skip_token(skipped_tokens,
                                                last_skipped_token_i, i, l)
                i += 1

            # Process year/month/day
            year, month, day = ymd.resolve_ymd(mstridx, yearfirst, dayfirst)
            if year is not None:
                res.year = year
                res.century_specified = ymd.century_specified

            if month is not None:
                res.month = month

            if day is not None:
                res.day = day

        except (IndexError, ValueError, AssertionError):
            return None, None

        if not info.validate(res):
            return None, None

        if fuzzy_with_tokens:
            return res, tuple(skipped_tokens)
        else:
            return res, None

    @staticmethod
    def _skip_token(skipped_tokens, last_skipped_token_i, i, l):
        if last_skipped_token_i == i - 1:
            # recombine the tokens
            skipped_tokens[-1] += l[i]
        else:
            # just append
            skipped_tokens.append(l[i])
        last_skipped_token_i = i
        return last_skipped_token_i


DEFAULTPARSER = parser()


def parse(timestr, parserinfo=None, **kwargs):
    """

    Parse a string in one of the supported formats, using the
    ``parserinfo`` parameters.

    :param timestr:
        A string containing a date/time stamp.

    :param parserinfo:
        A :class:`parserinfo` object containing parameters for the parser.
        If ``None``, the default arguments to the :class:`parserinfo`
        constructor are used.

    The ``**kwargs`` parameter takes the following keyword arguments:

    :param default:
        The default datetime object, if this is a datetime object and not
        ``None``, elements specified in ``timestr`` replace elements in the
        default object.

    :param ignoretz:
        If set ``True``, time zones in parsed strings are ignored and a naive
        :class:`datetime` object is returned.

    :param tzinfos:
            Additional time zone names / aliases which may be present in the
            string. This argument maps time zone names (and optionally offsets
            from those time zones) to time zones. This parameter can be a
            dictionary with timezone aliases mapping time zone names to time
            zones or a function taking two parameters (``tzname`` and
            ``tzoffset``) and returning a time zone.

            The timezones to which the names are mapped can be an integer
            offset from UTC in minutes or a :class:`tzinfo` object.

            .. doctest::
               :options: +NORMALIZE_WHITESPACE

                >>> from dateutil.parser import parse
                >>> from dateutil.tz import gettz
                >>> tzinfos = {"BRST": -10800, "CST": gettz("America/Chicago")}
                >>> parse("2012-01-19 17:21:00 BRST", tzinfos=tzinfos)
                datetime.datetime(2012, 1, 19, 17, 21, tzinfo=tzoffset(u'BRST', -10800))
                >>> parse("2012-01-19 17:21:00 CST", tzinfos=tzinfos)
                datetime.datetime(2012, 1, 19, 17, 21,
                                  tzinfo=tzfile('/usr/share/zoneinfo/America/Chicago'))

            This parameter is ignored if ``ignoretz`` is set.

    :param dayfirst:
        Whether to interpret the first value in an ambiguous 3-integer date
        (e.g. 01/05/09) as the day (``True``) or month (``False``). If
        ``yearfirst`` is set to ``True``, this distinguishes between YDM and
        YMD. If set to ``None``, this value is retrieved from the current
        :class:`parserinfo` object (which itself defaults to ``False``).

    :param yearfirst:
        Whether to interpret the first value in an ambiguous 3-integer date
        (e.g. 01/05/09) as the year. If ``True``, the first number is taken to
        be the year, otherwise the last number is taken to be the year. If
        this is set to ``None``, the value is retrieved from the current
        :class:`parserinfo` object (which itself defaults to ``False``).

    :param fuzzy:
        Whether to allow fuzzy parsing, allowing for string like "Today is
        January 1, 2047 at 8:21:00AM".

    :param fuzzy_with_tokens:
        If ``True``, ``fuzzy`` is automatically set to True, and the parser
        will return a tuple where the first element is the parsed
        :class:`datetime.datetime` datetimestamp and the second element is
        a tuple containing the portions of the string which were ignored:

        .. doctest::

            >>> from dateutil.parser import parse
            >>> parse("Today is January 1, 2047 at 8:21:00AM", fuzzy_with_tokens=True)
            (datetime.datetime(2047, 1, 1, 8, 21), (u'Today is ', u' ', u'at '))

    :return:
        Returns a :class:`datetime.datetime` object or, if the
        ``fuzzy_with_tokens`` option is ``True``, returns a tuple, the
        first element being a :class:`datetime.datetime` object, the second
        a tuple containing the fuzzy tokens.

    :raises ValueError:
        Raised for invalid or unknown string format, if the provided
        :class:`tzinfo` is not in a valid format, or if an invalid date
        would be created.

    :raises OverflowError:
        Raised if the parsed date exceeds the largest valid C integer on
        your system.
    """
    if parserinfo:
        return parser(parserinfo).parse(timestr, **kwargs)
    else:
        return DEFAULTPARSER.parse(timestr, **kwargs)


class _tzparser(object):

    class _result(_resultbase):

        __slots__ = ["stdabbr", "stdoffset", "dstabbr", "dstoffset",
                     "start", "end"]

        class _attr(_resultbase):
            __slots__ = ["month", "week", "weekday",
                         "yday", "jyday", "day", "time"]

        def __repr__(self):
            return self._repr("")

        def __init__(self):
            _resultbase.__init__(self)
            self.start = self._attr()
            self.end = self._attr()

    def parse(self, tzstr):
        res = self._result()
        l = _timelex.split(tzstr)
        try:

            len_l = len(l)

            i = 0
            while i < len_l:
                # BRST+3[BRDT[+2]]
                j = i
                while j < len_l and not [x for x in l[j]
                                         if x in "0123456789:,-+"]:
                    j += 1
                if j != i:
                    if not res.stdabbr:
                        offattr = "stdoffset"
                        res.stdabbr = "".join(l[i:j])
                    else:
                        offattr = "dstoffset"
                        res.dstabbr = "".join(l[i:j])
                    i = j
                    if (i < len_l and (l[i] in ('+', '-') or l[i][0] in
                                       "0123456789")):
                        if l[i] in ('+', '-'):
                            # Yes, that's right.  See the TZ variable
                            # documentation.
                            signal = (1, -1)[l[i] == '+']
                            i += 1
                        else:
                            signal = -1
                        len_li = len(l[i])
                        if len_li == 4:
                            # -0300
                            setattr(res, offattr, (int(l[i][:2])*3600 +
                                                   int(l[i][2:])*60)*signal)
                        elif i+1 < len_l and l[i+1] == ':':
                            # -03:00
                            setattr(res, offattr,
                                    (int(l[i])*3600+int(l[i+2])*60)*signal)
                            i += 2
                        elif len_li <= 2:
                            # -[0]3
                            setattr(res, offattr,
                                    int(l[i][:2])*3600*signal)
                        else:
                            return None
                        i += 1
                    if res.dstabbr:
                        break
                else:
                    break

            if i < len_l:
                for j in range(i, len_l):
                    if l[j] == ';':
                        l[j] = ','

                assert l[i] == ','

                i += 1

            if i >= len_l:
                pass
            elif (8 <= l.count(',') <= 9 and
                  not [y for x in l[i:] if x != ','
                       for y in x if y not in "0123456789"]):
                # GMT0BST,3,0,30,3600,10,0,26,7200[,3600]
                for x in (res.start, res.end):
                    x.month = int(l[i])
                    i += 2
                    if l[i] == '-':
                        value = int(l[i+1])*-1
                        i += 1
                    else:
                        value = int(l[i])
                    i += 2
                    if value:
                        x.week = value
                        x.weekday = (int(l[i])-1) % 7
                    else:
                        x.day = int(l[i])
                    i += 2
                    x.time = int(l[i])
                    i += 2
                if i < len_l:
                    if l[i] in ('-', '+'):
                        signal = (-1, 1)[l[i] == "+"]
                        i += 1
                    else:
                        signal = 1
                    res.dstoffset = (res.stdoffset+int(l[i]))*signal
            elif (l.count(',') == 2 and l[i:].count('/') <= 2 and
                  not [y for x in l[i:] if x not in (',', '/', 'J', 'M',
                                                     '.', '-', ':')
                       for y in x if y not in "0123456789"]):
                for x in (res.start, res.end):
                    if l[i] == 'J':
                        # non-leap year day (1 based)
                        i += 1
                        x.jyday = int(l[i])
                    elif l[i] == 'M':
                        # month[-.]week[-.]weekday
                        i += 1
                        x.month = int(l[i])
                        i += 1
                        assert l[i] in ('-', '.')
                        i += 1
                        x.week = int(l[i])
                        if x.week == 5:
                            x.week = -1
                        i += 1
                        assert l[i] in ('-', '.')
                        i += 1
                        x.weekday = (int(l[i])-1) % 7
                    else:
                        # year day (zero based)
                        x.yday = int(l[i])+1

                    i += 1

                    if i < len_l and l[i] == '/':
                        i += 1
                        # start time
                        len_li = len(l[i])
                        if len_li == 4:
                            # -0300
                            x.time = (int(l[i][:2])*3600+int(l[i][2:])*60)
                        elif i+1 < len_l and l[i+1] == ':':
                            # -03:00
                            x.time = int(l[i])*3600+int(l[i+2])*60
                            i += 2
                            if i+1 < len_l and l[i+1] == ':':
                                i += 2
                                x.time += int(l[i])
                        elif len_li <= 2:
                            # -[0]3
                            x.time = (int(l[i][:2])*3600)
                        else:
                            return None
                        i += 1

                    assert i == len_l or l[i] == ','

                    i += 1

                assert i >= len_l

        except (IndexError, ValueError, AssertionError):
            return None

        return res


DEFAULTTZPARSER = _tzparser()


def _parsetz(tzstr):
    return DEFAULTTZPARSER.parse(tzstr)


def _parsems(value):
    """Parse a I[.F] seconds value into (seconds, microseconds)."""
    if "." not in value:
        return int(value), 0
    else:
        i, f = value.split(".")
        return int(i), int(f.ljust(6, "0")[:6])


# vim:ts=4:sw=4:et



from unittest import TestCase

import simplejson as json


class TestBitSizeIntAsString(TestCase):
    # Python 2.5, at least the one that ships on Mac OS X, calculates
    # 2 ** 31 as 0! It manages to calculate 1 << 31 correctly.
    values = [
        (200, 200),
        ((1 << 31) - 1, (1 << 31) - 1),
        ((1 << 31), str(1 << 31)),
        ((1 << 31) + 1, str((1 << 31) + 1)),
        (-100, -100),
        ((-1 << 31), str(-1 << 31)),
        ((-1 << 31) - 1, str((-1 << 31) - 1)),
        ((-1 << 31) + 1, (-1 << 31) + 1),
    ]

    def test_invalid_counts(self):
        for n in ['foo', -1, 0, 1.0]:
            self.assertRaises(
                TypeError,
                json.dumps, 0, int_as_string_bitcount=n)

    def test_ints_outside_range_fails(self):
        self.assertNotEqual(
            str(1 << 15),
            json.loads(json.dumps(1 << 15, int_as_string_bitcount=16)),
            )

    def test_ints(self):
        for val, expect in self.values:
            self.assertEqual(
                val,
                json.loads(json.dumps(val)))
            self.assertEqual(
                expect,
                json.loads(json.dumps(val, int_as_string_bitcount=31)),
                )

    def test_lists(self):
        for val, expect in self.values:
            val = [val, val]
            expect = [expect, expect]
            self.assertEqual(
                val,
                json.loads(json.dumps(val)))
            self.assertEqual(
                expect,
                json.loads(json.dumps(val, int_as_string_bitcount=31)))

    def test_dicts(self):
        for val, expect in self.values:
            val = {'k': val}
            expect = {'k': expect}
            self.assertEqual(
                val,
                json.loads(json.dumps(val)))
            self.assertEqual(
                expect,
                json.loads(json.dumps(val, int_as_string_bitcount=31)))

    def test_dict_keys(self):
        for val, _ in self.values:
            expect = {str(val): 'value'}
            val = {val: 'value'}
            self.assertEqual(
                expect,
                json.loads(json.dumps(val)))
            self.assertEqual(
                expect,
                json.loads(json.dumps(val, int_as_string_bitcount=31)))



class CleanDatabasesRequest(object):
    def __init__(
        self,
        default_outdated_events_days,  # type: int
    ):
        # type: (...) -> None

        self.default_outdated_events_days = default_outdated_events_days



# encoding: utf-8

from justificationdiscountreport import JustificationDiscountReport
from justificationdiscountreport.repository import JustificationDiscountRepository
from report import Generator
from reportsdata.justificationdiscountreport.dto import (
    JustificationDiscountBodyDto,
    JustificationDiscountHeaderDto,
    JustificationDiscountGeneratorDto,
)


class JustificationDiscountReportGenerator(Generator):

    def __init__(
        self,
        repository,         # type: JustificationDiscountRepository
        initial_date,       # type: str
        end_date,           # type: str
        operator_id,        # type: str
        pos_id,             # type: str
    ):
        # type: (...) -> None

        self.repository = repository
        self.initial_date = initial_date
        self.end_date = end_date
        self.operator_id = operator_id
        self.pos_id = pos_id

    def generate_data(
        self,
    ):
        # type: (...) -> JustificationDiscountGeneratorDto

        justification_discount_report = JustificationDiscountReport(
            justification_discount_repository=self.repository,
        )

        justifications = justification_discount_report.get_justifications(
            initial_date=self.initial_date,
            end_date=self.end_date,
            operator_id=self.operator_id,
        )

        header = JustificationDiscountHeaderDto(
            operator=self.operator_id,
            start_date=self.initial_date,
            end_date=self.end_date,
            pos_id=self.pos_id,
        )

        body = JustificationDiscountBodyDto(
            justifications=justifications,
        )

        return JustificationDiscountGeneratorDto(
            header=header,
            body=body,
        )



# encoding: utf-8

from old_helper import BaseRepository


class TableServiceRepository(BaseRepository):

    def __init__(self, mbcontext):
        super(TableServiceRepository, self).__init__(mbcontext)

    def get_paid_ticket(self, business_period):
        def inner_func(conn):
            query = self._PaidTicketsQuery.format(business_period)
            users_average_ticket = [(x.get_entry(0), x.get_entry(1)) for x in conn.select(query)]
            return users_average_ticket
        return self.execute_with_connection(inner_func)

    _PaidTicketsQuery = """\
SELECT UserId, SUM(NumberOfSeats) AS seats
FROM TableService tb
WHERE BusinessPeriod = '{}' AND (OrderId IS NOT NULL OR TotalAmount IS NOT NULL)
GROUP BY UserId
"""



from ._CustomExceptions import (
    AuthValidationErrors,
    BadRequestError,
    ClientCustomError,
    ExternalServiceErrors,
    InUseError,
    NotAvailableError,
    NotFoundError,
    OperationErrors,
    ValidationError,
)



import logging
import sysactions

from sysactions import mbcontext
from typing import Optional

from tef import PrintReceipt

logger = logging.getLogger("PosActions")


def reprint_receipt(
    pos_id,                 # type: str
    receipt,                # type: str
    fiscalization_type="",  # type: Optional[str]
):
    # type: (...) -> bool

    eft_check = _check_eft_value(
        pos_id=pos_id,
        eft_value=receipt,
        error_message="$PAYMENT_HAS_NO_TEF",
    )
    if not eft_check:
        return False

    printer_receipt = PrintReceipt(
        mb_context=mbcontext,
        logger=logger,
    )

    reprinted = printer_receipt.reprint_receipt(
        pos_id=pos_id,
        receipt=receipt,
        fiscalization_type=fiscalization_type,
    )

    return reprinted


def _check_eft_value(
    pos_id,         # type: str
    eft_value,      # type: str
    error_message,  # type: str
):
    # type (...) -> bool

    if eft_value in ['', None]:
        sysactions.show_messagebox(
            posid=pos_id,
            message=error_message,
            title="$REPRINT_TEF_ERROR",
            buttons="$OK",
        )
        return False

    return True



from xml.etree import cElementTree as eTree

from application.interactor import OrderInteractor

NOT_ELECTRONIC_TYPE_ID = (0, 9)


def cancel_non_eft_transactions(
    pos_id,     # type: str
    order_id,   # type: str
    order,      # type: eTree.Element
):
    # type: (...) -> None

    tenders = order.findall("TenderHistory/Tender")
    non_electronic_tenders = [
        tender
        for tender in tenders
        if int(tender.get("electronicType") or tender.get("eletronicType")) in NOT_ELECTRONIC_TYPE_ID
    ]

    order_interactor = OrderInteractor(pos_id=pos_id)
    for tender in non_electronic_tenders:
        tender_amount = float(tender.get('tenderAmount'))
        if tender_amount == 0:
            continue

        order_interactor.clear_order_tender(
            order_tender_id=tender.get("tenderId"),
            order_id=order_id or order.get("orderId"),
        )



import json
from xml.etree import cElementTree as eTree

from typing import Optional

from application.service import OrderService

from application.interactor import OnOrderReopenInteractor
from application.model import Configurations
from messagebus import DefaultToken, DataType, Message
from messageprocessorutil import UuidMessageProcessor


class OnOrderReopenProcessor(UuidMessageProcessor):
    def __init__(self, configs, interactor, order_service):
        # type: (Configurations, OnOrderReopenInteractor, OrderService) -> None

        super(OnOrderReopenProcessor, self).__init__()
        self.logger = configs.logger
        self.configs = configs
        self.interactor = interactor
        self.order_service = order_service

    def parse_data(self, data, event_type):
        # type: (str, str) -> int

        data = json.loads(data)
        pos_id = int(data.get("pos_id"))

        return pos_id

    def call_business(self, pos_id):
        # type: (int) -> Optional[eTree.XML]

        return self.interactor.execute(pos_id)

    def format_response(self, message_bus, message, event, request, response):
        if message:
            data = eTree.tostring(element=response) if response is not None else ""
            message_bus.reply_message(
                message=message,
                reply_message=Message(
                    token=DefaultToken.TK_SYS_ACK,
                    data_type=DataType.string,
                    data=data,
                ),
            )

    def format_exception(self, message_bus, message, event, input_obj, exception):
        if message:
            message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))

    def format_parse_exception(self, message_bus, message, event, data, exception):
        if message:
            message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))
